import sys
import json
import argparse
from typing import Any
from .cropper import Cropper

class ArgumentParserWithConfig(argparse.ArgumentParser):
    """An ArgumentParser that loads default values from a config file.

    This class extends the :class:`argparse.ArgumentParser` class to 
    load default values from a config file specified by a command-line 
    argument.
    """
    def __init__(
        self,
        *args,
        config_arg: str | list[str] = ["-c", "--config"],
        **kwargs
    ):
        """Initialize ArgumentParserWithConfig object.

        Args:
            *args: Additional arguments for initializing 
                :class:`argparse.ArgumentParser`.
            config_arg: The name (or the list of names) of the 
                command-line argument that specifies the path to the 
                JSON config file. Defaults to ["-c", "--config"].
            **kwargs: Additional keyword arguments for initializing 
                :class:`argparse.ArgumentParser`.
        """
        super().__init__(*args, **kwargs)
        self.config_arg = config_arg

        if isinstance(self.config_arg, str):
            # Convert string to list of strings
            self.config_arg = [self.config_arg]

        self.add_argument(
        *config_arg, type=str, 
        help=f"Path to JSON file with arguments. If other arguments are "
             f"further specified via command line, they will overwrite the "
             f"ones with the same name in the JSON file.")

    def parse_args(
        self, 
        args: Any | None = None, 
        namespace: argparse.Namespace | None = None
    ) -> argparse.Namespace:
        """Parse arguments and load default values from the config file.

        This method parses the command-line arguments and loads default 
        values from the config file specified by the ``config_arg`` 
        parameter. The default values are only loaded once.

        Args:
            args: The sequence of arguments to parse. If None, then the 
                sequence will be retrieved by reading the command-line 
                arguments. Defaults to None.
            namespace: The namespace object to extend with parsed 
                arguments. If None, then a new object will be created. 
                Defaults to None.

        Returns:
            A namespace object containing the parsed arguments.
        """
        # Either load the sequence as a list or read command-line
        args = sys.argv[1:] if args is None else list(args)
        
        if len(cfg := (set(self.config_arg)) & set(args)) > 0:
            # Pop the config key and the value in the arg list
            args.pop(index := args.index(next(iter(cfg))))
            config_path = args.pop(index)

            with open(config_path) as f:
                # Load new default val dict
                new_defaults = json.load(f)

            for key, val in new_defaults.items():
                for action in self._actions:
                    if key == action.dest \
                       and action.default is not argparse.SUPPRESS:
                        # Update default val
                        action.default = val
                        break

        for action in self._actions:
            if set(action.option_strings) == set(self.config_arg):
                # Remove config parse action
                self._remove_action(action)
                break

        # Parse the remaining arguments (no config)
        args = super().parse_args(args, namespace)

        return args

def parse_args() -> dict[str, Any]:
    """Parses command-line arguments.

    Defines the possible command-line arguments that must match the 
    acceptable arguments by :class:`.Cropper`. It parses those arguments 
    and converts them to a dictionary.

    Raises:
        ValueError: If ``input_dir`` is not specified.

    Returns:
        A dictionary where keys represent argument names and values 
        represent those argument values.
    """
    # Arg parser that can load default values
    parser = ArgumentParserWithConfig()

    # TODO: check list args, check transparent, check attr_groups
    parser.add_argument(
        "-i", "--input_dir", type=str, 
        help="Path to input directory with image files.")
    parser.add_argument(
        "-o", "--output-dir", type=str, 
        help=f"Path to output directory to save the extracted face images. If "
             f"not specified, the same path is used as for input_dir, except "
             f"'_faces' suffix is added the name.")
    parser.add_argument(
        "-s", "--output-size", type=int, nargs='+', default=[256, 256], 
        help=f"The output size (width, height) of cropped image faces. If "
             f"provided as a single number, the same value is used for both "
             f"width and height. Defaults to [256, 256].")
    parser.add_argument(
        "-f", "--output-format", type=str, 
        help=f"The output format of the saved face images, e.g., 'jpg', 'png'."
             f" If not specified, the same format as the image from which the "
             f"face is extracted will be used.")
    parser.add_argument(
        "-r", "--resize-size", type=int, nargs='+', default=[1024, 1024], 
        help=f"The interim size (width, height) each image should be resized "
             f"to before processing them. If provided as a single number, the "
             f"same value is used for both width and height. Defaults to "
             f"[1024, 1024].")
    parser.add_argument(
        "-ff", "--face-factor", type=float, default=0.65, 
        help=f"The fraction of the face area relative to the output image. "
             f"Defaults to 0.65.")
    parser.add_argument(
        "-st", "--strategy", type=str, default="largest",
        choices=["all", "best", "largest"],
        help=f"The strategy to use to extract faces from each image. Defaults "
             f"to 'largest'.")
    parser.add_argument(
        "-p", "--padding", type=str, default="reflect",
        choices=["constant", "replicate", "reflect", "wrap", "reflect_101"], 
        help=f"The padding type (border mode) to apply when cropping out faces"
             f" near edges. Defaults to 'reflect'.")
    parser.add_argument(
        "-a", "--allow-skew", action="store_true", 
        help=f"Whether to allow skewing the faces to better match the the "
             f"standard (average) face landmarks.")
    parser.add_argument(
        "-l", "--landmarks", type=str, 
        help=f"Path to landmarks file if landmarks are already known and "
             f"prediction is not needed. Common file types are json "
             f"(\"image.jpg\": [x1, y1, ...]), csv (image.jpg,x1,y1,...; "
             f"first line is header), txt and other (image.jpg x1 y2).")
    parser.add_argument(
        "-ag", "--attr-groups", type=json.loads,
        help=f"Attribute groups dictionary that specifies how to group the "
             f"output face images according to some common attributes. Should "
             f"be a JSON-parsable string dictionary of type "
             f"dict[str, list[int]], e.g., '{{\"glasses\": [6]}}'.")
    parser.add_argument(
        "-mg", "--mask-groups", type=json.loads, 
        help=f"Mask groups dictionary that specifies how to group the output "
             f"face images according to some face attributes that make up a "
             f"segmentation mask. Should be a JSON-parsable string dictionary "
             f"of type dict[str, list[int]], e.g., '{{\"eyes\": [4, 5]}}'.")
    parser.add_argument(
        "-dt", "--det-threshold", type=float, default=0.6, 
        help=f"The visual threshold, i.e., minimum confidence score, for a "
             f"detected face to be considered an actual face. If a negative "
             f"value is provided, e.g., -1, landmark prediction is not "
             f"performed. Defaults to 0.6.")
    parser.add_argument(
        "-et", "--enh-threshold", type=float, default=0.001, 
        help=f"Quality enhancement threshold that tells when the image quality"
             f"should be enhanced. It is the minimum average face factor in "
             f"the input image. Defaults to 0.001.")
    parser.add_argument(
        "-b", "--batch-size", type=int, default=8, 
        help=f"The batch size. It is the maximum number of images that can be "
             f"processed by every processor at a single time-step. Defaults "
             f"to 8.")
    parser.add_argument(
        "-n", "--num-processes", type=int, default=1, 
        help=f"The number of processes to launch to perform image processing. "
             f"If landmarks are provided and no quality enhancement or "
             f"attribute grouping is done, feel free to set this to the "
             f"number of CPUs your machine has. Defaults to 1.")
    parser.add_argument(
        "-d", "--device", type=str, default="cpu", 
        help=f"The device on which to perform the predictions, i.e., landmark "
             f"detection, quality enhancement and face  parsing. Defaults to "
             f"'cpu'.")

    # Parse arguments and convert to dict
    kwargs = vars(parser.parse_args())
    
    if kwargs["input_dir"] is None:
        raise ValueError("Input directory must be specified.")
    
    if kwargs["det_threshold"] is not None and kwargs["det_threshold"] < 0:
        # If negative, set it to None
        kwargs["det_threshold"] = None
    
    if kwargs["enh_threshold"] is not None and kwargs["enh_threshold"] < 0:
        # If negative, set it to None
        kwargs["enh_threshold"] = None
    
    return kwargs
    

def main():
    """Processes an input dir of images

    Creates a cropper object based on the provided command-line 
    arguments and processes a specified directory of images. There are 
    3 main features the cropper can do (either together or separately): 
    align and center-crop face images, enhance quality, group by 
    attributes. For more details, see :class:`.Cropper`.
    """
    # Parse arguments
    kwargs = parse_args()

    # Pop the input and output dirs
    input_dir = kwargs.pop("input_dir")
    output_dir = kwargs.pop("output_dir")

    # Init cropper and process
    cropper = Cropper(**kwargs)
    cropper.process_dir(input_dir, output_dir)

if __name__ == "__main__":
    main()


import os
import cv2
import json
import torch
import warnings
import numpy as np


STANDARD_LANDMARKS_5 = np.float32([
    [0.31556875000000000, 0.4615741071428571],
    [0.68262291666666670, 0.4615741071428571],
    [0.50026249999999990, 0.6405053571428571],
    [0.34947187500000004, 0.8246919642857142],
    [0.65343645833333330, 0.8246919642857142],
])

def parse_landmarks_file(
    file_path: str,
    **kwargs,
) -> tuple[np.ndarray, np.ndarray]:
    """Parses landmarks file.

    Reads the file containing landmark coordinates and corresponding 
    image file names and generates a numpy array of those names and a 
    corresponding numpy array of those sets of landmarks.

    The files are expected to be formatted as follows:

        * `.json`::

            {
                "image_1.jpg": [23, 45, 64, 47, ...],
                "image_2.jpg": [17, 32, 30, 29, ...],
                ...
            }

        * `.csv`::

            images,x1,y1,x2,y2,...
            image_1.jpg,23,45,64,47,...
            image_2.jpg,17,32,30,29,...
            ...

        * `.txt` and other::

            image_1.jpg 23 45 64 47 ...
            image_2.jpg 17 32 30 29 ...
            ...

    Note:
        The number of landmarks does not matter, all will be 
        transformed to shape (-1, 2), where -1 stands for the number of 
        facial points (landmark coordinates), e.g., 5, 68 etc.

    Args:
        file_path: The path to the landmarks file.
        **kwargs: Additional keyword arguments that go into 
            :func:`numpy.genfromtxt`. Please do not specify *dtype* and 
            *usecols* arguments as they are already specified.

    Returns:
        A tuple where the first element is the parsed landmarks batch as 
        a numpy array of shape (N, ``num_landm``, 2) of type 
        :attr:`numpy.float32` and  the second element is a corresponding 
        batch of image file names of shape (N,) of type 
        :class:`numpy.str_`.
    """
    if file_path.endswith(".json"):
        with open(file_path, 'r') as f:
            # Read and parse
            data = json.load(f)
            filenames = np.array(data.keys())
            landmarks = np.array(data.values())
    else:
        if file_path.endswith(".csv"):
            # Set default params for csv files
            kwargs.setdefault("delimiter", ',')
            kwargs.setdefault("skip_header", 1)

        # Use the first column for filenames, the rest for landmarks
        filenames = np.genfromtxt(file_path, usecols=0, dtype=str, **kwargs)
        landmarks = np.genfromtxt(file_path, dtype=np.float32, **kwargs)[:, 1:]

    return landmarks.reshape(len(landmarks), -1, 2), filenames

def get_landmark_slices_5(num_landmarks: int) -> list[slice]:
    """Gets the landmarks slices that show where the 5 landmarks are.

    Generates slices of which coordinates to select in a larger set of 
    landmarks (e.g., 12, 68, 106) to represent the coordinates of 
    5-points landmarks set. The slice of indices can be used to select 
    multiple coordinates and average them to a single point.

    Args:
        num_landmarks: The number of landmarks in the larger set.

    Raises:
        ValueError: If the number of landmarks in the larger set is 
            not supported.

    Returns:
        A list of slices where each slice indicates the indices of 
        coordinates to select from the larger set of landmarks to 
        represent a 5-point landmarks set.
    """
    match num_landmarks:
        case 5:
            indices = [(0, 1), (1, 2), (2, 3), (3, 4), (4, 5)]
        case 12:
            indices = [(10, 11), (11, 12), (2, 3), (3, 4), (4, 5)]
        case 17:
            indices = [(2, 5), (7, 10), (10, 11), (13, 14), (16, 17)]
        case 21:
            indices = [(6, 9), (9, 12), (14, 15), (17, 18), (19, 20)]
        case 29:
            indices = [(4, 9), (13, 18), (19, 20), (22, 23), (27, 28)]
        case 49: # same as 51
            indices = [(19, 25), (25, 31), (13, 14), (31, 32), (37, 38)]
        case 68:
            indices = [(36, 42), (42, 48), (30, 31), (48, 49), (54, 55)]
        case 98:
            indices = [(60, 68), (68, 76), (54, 55), (76, 77), (82, 83)]
        case 106:
            indices = [(66, 75), (75, 84), (54, 55), (85, 86), (91, 92)]
        case _:
            raise ValueError(f"Invalid number of landmarks: {num_landmarks}")

    return [slice(*x) for x in indices]

def get_ldm_slices(
    num_tgt_landmarks: int,
    num_src_landmarks: int,
) -> list[slice]:
    """Generates a list of slices that form a reduced landmarks set.

    Takes the number of target landmarks and the number of source 
    landmarks and generates slices that show which coordinates to select 
    from a larger landmarks set (that should be averaged) to form a 
    reduced landmarks set that has the same number of landmarks as the 
    target landmarks set.

    Args:
        num_tgt_landmarks: The number of reduced landmarks to generate 
            slices for.
        num_src_landmarks: The number of actual landmarks that is larger 
            or equal to the number of target landmarks. Based on this 
            number, generated slices will contain different indices.

    Raises:
        ValueError: If the number of target landmarks is not supported.

    Returns:
        A list of slices where each slice indicates the indices of 
        coordinates to select from the larger set of landmarks to 
        represent a reduced (like target) landmarks set.
    """
    match num_tgt_landmarks:
        case 5:
            slices = get_landmark_slices_5(num_src_landmarks)
        case _:
            raise ValueError(f"The number of target (standard) landmarks is "
                             f"not supported {num_tgt_landmarks}")
    
    return slices

def as_numpy(
    img: torch.Tensor | np.ndarray | list[torch.Tensor] | list[np.ndarray],
) -> np.ndarray | list[np.ndarray]:
    """Converts batch of images to numpy type.

    Converts a batch of images to numpy type. UINT8 type and channel 
    dimension is last. If the batch of images is already of numpy type, 
    it is simply returned.

    Args:
        img: The image batch represented as a torch tensor of shape 
            (N, 3, H, W) or a list of torch tensors of different 
            spatial sizes.

    Returns:
        A batch of images represented as a numpy array of shape 
        (N, H, W, 3) of type :attr:`numpy.uint8` or a list of numpy 
        arrays of different spatial sizes.
    """
    if isinstance(img[0], np.ndarray):
        return img
    elif isinstance(img, list):
        img = [x.permute(1, 2, 0).cpu().numpy().astype(np.uint8) for x in img]
    else:
        img = img.permute(0, 2, 3, 1).cpu().numpy().astype(np.uint8)

    return img

def as_tensor(
    img: np.ndarray | torch.Tensor | list[np.ndarray] | list[torch.Tensor],
    device: str | torch.device = "cpu",
) -> torch.Tensor | list[torch.Tensor]:
    """Converts batch of images to torch tensor type.

    Converts a batch of images to torch tensor type. Float 32 type and 
    channel dimension is before spatial dimension. If the batch of 
    images is already of torch tensor type, it is simply returned.

    Args:
        img: The image batch represented as a numpy array of shape 
            (N, H, W, 3) or a list of numpy arrays of different 
            spatial sizes.
        device: The device on which to return the torch tensor.

    Returns:
        A batch of images represented as a torch tensor of shape 
        (N, 3, H, W) of type :attr:`torch.float32` or a list of torch 
        tensors of different spatial sizes.
    """
    if isinstance(img[0], torch.Tensor):
        return img
    elif isinstance(img, list):
        img = [torch.from_numpy(x).permute(2, 0, 1).float().to(device) for x in img]
    else:
        img = torch.from_numpy(img).permute(0, 3, 1, 2).float().to(device)
    
    return img

def read_images(
    file_names: list[str],
    input_dir: str,
) -> tuple[list[np.ndarray], np.ndarray]:
    """Reads images from the specified paths.

    Takes a list of file names and an input directory and loads the 
    specified images from it. If an image could not be loaded, a warning 
    is raised.

    Args:
        file_names: The list of image file names.
        input_dir (str): The input directory with the images.

    Returns:
        A tuple where the first element is a list of length N (number of 
        loaded images) where each element is an RGB image represented as 
        a numpy array of type :attr:`numpy.uint8` of shape (H, W, 3) 
        (note that H and W for each image can be different) and the 
        second element is a numpy array of type :class:`numpy.str_` of 
        shape (N,) representing the list of file names that correspond 
        to each image. The second element is just a numpy form of 
        ``file_names`` but it can have a lower length, in case some 
        images were not read successfully.
    """
    # Init index and img list
    indices, images = [], []

    for i, file_name in enumerate(file_names):
        # Generate full path to the input image
        path = os.path.join(input_dir, file_name)

        try:
            # Read the image from the given path, convert to RGB form
            image = cv2.cvtColor(cv2.imread(path), cv2.COLOR_BGR2RGB)
        except cv2.error as e:
            warnings.warn(f"Could not read the image {path}")
            continue
        
        # Add image and index
        images.append(image)
        indices.append(i)
    
    return images, np.array(file_names)[indices]

def as_batch(
    images: list[np.ndarray],
    size: int | tuple[int, int] = 512,
    padding_mode: str = "constant",
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Creates image batch from a list of image paths.

    For every image in the given list, the image is resized to not 
    exceed either of the dimensions specified in ``size`` while keeping 
    the same aspect ratio and the shorter dimension is padded to fully 
    match the specified size. All the images are stacked and returned as 
    a batch. Variables required to transform the images back to the 
    original ones (padding and scale) are also returned as a batch.

    Example:
        If some loaded image dimension is (1280×720) and the desired
        output ``size`` is specified as *(512, 256)*, then the image is
        first be resized to *(455, 256)* and then the width is padded 
        from both sides. The final image size is *(512, 256)*.

    Args:
        images: The list of paths to images.
        padding_mode: The type of padding to apply to pad the shorter
            dimension. For the available options, see
            `OpenCV BorderTypes <https://docs.opencv.org/3.4/d2/de8/group__core__array.html#ga209f2f4869e304c82d07739337eae7c5>`_.
            It can be all lowercase. Defaults to "constant".
        size: The width and the height each image should be resized +
            padded to. I.e., the spacial dimensions of the batch. If
            a single number is specified then it is the same for width
            height. Defaults to 512.

    Returns:
        A tuple of stacked numpy arrays representing 3 batches - resized 
        + padded images of shape (N, H, W, 3) of type 
        :attr:`numpy.uint8` with values from 0 to 255, un-scale factors 
        of shape (N,) of type :attr:`numpy.float32`, and applied 
        paddings of shape (N, 4) of type :attr:`numpy.int64` with 
        values >= 0.
    """
    # Init lists, resize dims, border type
    img_batch, unscales, paddings = [], [], []
    size = (size, size) if isinstance(size, int) else size
    border_type = getattr(cv2, f"BORDER_{padding_mode.upper()}")

    for image in images:    
        # Get width, height, padding & check interpolation
        (h, w), m = image.shape[:2], max(*image.shape[:2])
        interpolation = cv2.INTER_AREA if m > max(size) else cv2.INTER_CUBIC

        if (ratio_w := size[0] / w) < (ratio_h := size[1] / h):
            # Based on width 
            unscale = ratio_w
            (ww,hh) = size[0], int(h * ratio_w)
            padding = [(size[1] - hh) // 2, (size[1] - hh + 1) // 2, 0, 0]
        else:
            # Based on height
            unscale = ratio_h
            (ww,hh) = int(w * ratio_h), size[1]
            padding = [0, 0, (size[0] - ww) // 2, (size[0] - ww + 1) // 2]
    
        # Pad the lower dimension with specific border type, then resize
        image = cv2.resize(image, (ww, hh), interpolation=interpolation)
        image = cv2.copyMakeBorder(image, *padding, borderType=border_type)

        # Append to lists
        img_batch.append(image)
        unscales.append(np.array(unscale))
        paddings.append(np.array(padding))

    return np.stack(img_batch), np.stack(unscales), np.stack(paddings)



import os
import cv2
import tqdm
import torch
import numpy as np

from functools import partial
from collections import defaultdict
from multiprocessing.pool import ThreadPool

from .models import BiSeNet
from .models import RRDBNet
from .models import RetinaFace

from .utils import (
    STANDARD_LANDMARKS_5, 
    parse_landmarks_file, 
    get_ldm_slices, 
    as_numpy, 
    as_tensor,
    read_images,
    as_batch,
)


class Cropper():
    """Face cropper class with bonus features.

    This class is capable of automatically aligning and center-cropping 
    faces, enhancing image quality and grouping the extracted faces 
    according to specified face attributes, as well as generating masks 
    for those attributes.
    
    Capabilities
    ------------

    This class has the following 3 main features:

        1. **Face cropping** - automatic face alignment and cropping 
           based on landmarks. The landmarks can either be predicted via 
           face detection model (see :class:`.RetinaFace`) or they 
           can be provided as txt, csv, json etc. file. It is possible 
           to control face factor in the extracted images and strategy 
           of extraction (e.g., largest face, all faces per image).
        2. **Face enhancement** - automatic quality enhancement of 
           images where the relative face area is small. For instance, 
           there may be images with many faces, but the quality of those 
           faces, if zoomed in, is low. Quality enhancement feature 
           allows to remove the blurriness. It can also enhance the 
           quality of every image, if desired (see 
           :class:`.RRDBNet`).
        3. **Face parsing** - automatic face attribute parsing and 
           grouping to sub-directories according selected attributes. 
           Attributes can indicate to group faces that contain specific 
           properties, e.g., "earrings and neckless", "glasses". They 
           can also indicate what properties the faces should not 
           include to form a group, e.g., "no accessories" group would 
           indicate to include faces without hats, glasses, earrings, 
           neckless etc. It is also possible to generate masks for 
           selected face attributes, e.g., "glasses", 
           "eyes and eyebrows". For more intuition on how grouping 
           works, see :class:`.BiSeNet` and 
           :meth:`save_groups`.

    The class is designed to perform all or some combination of the 
    functions in one go, however, each feature is independent of one 
    another and can work one by one. For example, it is possible to 
    first extract all the faces in some output directory, then apply 
    quality enhancement for every face to produce better quality faces 
    in another output directory and then apply face parsing to group 
    faces into different sub-folders according to some common attributes 
    in a final output directory.

    It is possible to configure the number of processing units and the 
    batch size for significant speedups., if the hardware allows.

    Examples
    --------

    Command line example
        >>> python face_crop_plus -i path/to/images -o path/to/out/dir

    Auto face cropping (with face factor) and quality enhancement:
        >>> cropper = Cropper(face_factor=0.7, enh_threshold=0.01)
        >>> cropper.process_dir(input_dir="path/to/images")

    Very fast cropping with already known landmarks (no enhancement):
        >>> cropper = Cropper(landmarks="path/to/landmarks.txt", 
                              num_processes=24,
                              enh_threshold=None)
        >>> cropper.process_dir(input_dir="path/to/images")

    Face cropping to attribute groups to custom output dir:
        >>> attr_groups = {"glasses": [6], "no_glasses_hats": [-6, -18]}
        >>> cropper = Cropper(attr_groups=attr_groups)
        >>> inp, out = "path/to/images", "path/to/parent/out/dir"
        >>> cropper.process_dir(input_dir=inp, output_dir=out)

    Face cropping and grouping by face attributes (+ generating masks):
        >>> groups = {"glasses": [6], "eyes_and_eyebrows": [2, 3, 4, 5]}
        >>> cropper = Cropper(output_format="png", mask_groups=groups)
        >>> cropper.process_dir("path/to/images")

    For grouping by face attributes, see documented face attribute 
    indices in :class:`.BiSeNet`.

    Class Attributes
    ----------------

    For how to initialize the class and to understand its functionality 
    better, please refer to class attributes initialized via 
    :meth:`__init__`. Here, further class attributes are 
    described automatically initialized via  :meth:`_init_models` and 
    :meth:`_init_landmarks_target`.

    Attributes:
        det_model (RetinaFace): Face detection model 
            (:class:`torch.nn.Module`) that is capable of detecting 
            faces and predicting landmarks used for face alignment. See 
            :class:`.RetinaFace`.
        enh_model (RRDBNet): Image quality enhancement model 
            (torch.nn.Module) that is capable of enhancing the quality 
            of images with faces. It can automatically detect which 
            faces to enhance based on average face area in the image, 
            compared to the whole image area. See :class:`.RRDBNet`.
        par_model (BiSeNet): Face parsing model (torch.nn.Module) that 
            is capable of classifying pixels according to specific face 
            attributes, e.g., "left_eye", "earring". It is able to group 
            faces to different groups and generate attribute masks. See 
            :class:`.BiSeNet`.
        landmarks_target (numpy.ndarray): Standard normalized landmarks 
            of shape  (``self.num_std_landmarks``, 2). These are scaled 
            by ``self.face_factor`` and used as ideal landmark 
            coordinates for the extracted faces. In other words, they 
            are reference landmarks used to estimate the transformation 
            of an image based on some actual set of face landmarks for 
            that image.
    """
    def __init__(
        self,
        output_size: int | tuple[int, int] | list[int] = 256,
        output_format: str | None = None,
        resize_size: int | tuple[int, int] | list[int] = 1024,
        face_factor: float = 0.65,
        strategy: str = "largest",
        padding: str = "reflect",
        allow_skew: bool = False,
        landmarks: str | tuple[np.ndarray, np.ndarray] | None = None,
        attr_groups: dict[str, list[int]] | None = None,
        mask_groups: dict[str, list[int]] | None = None,
        det_threshold: float | None = 0.6,
        enh_threshold: float | None = 0.001,
        batch_size: int = 8,
        num_processes: int = 1,
        device: str | torch.device = "cpu",
    ):
        """Initializes the cropper.

        Initializes class attributes. 

        Args:
            output_size: The output size (width, height) of cropped 
                image faces. If provided as a single number, the same 
                value is used for both width and height. Defaults to
                256.
            output_format: The output format of the saved face images. 
                For available options, see 
                `OpenCV imread <https://docs.opencv.org/4.x/d4/da8/group__imgcodecs.html#ga288b8b3da0892bd651fce07b3bbd3a56>`_. 
                If not specified, then the same image extension will not 
                be changed, i.e., face images will be of the same format 
                as the images from which they are extracted. Defaults to 
                None.
            resize_size: The interim size (width, height) each image 
                should be resized to before processing images. This is 
                used to resize images to a common size to allow to make 
                a batch. It should ideally be the mean width and height 
                of all the images to be processed (but can simply be a
                square). Images will be resized to to the specified size 
                while maintaining the aspect ratio (one of the 
                dimensions will always match either the specified width 
                or height). The shorter dimension would afterwards be 
                padded - for more information on how it works, see 
                :func:`.utils.create_batch_from_files`. Defaults to 
                1024.
            face_factor: The fraction of the face area relative to the 
                output image. Defaults to 0.65.
            strategy: The strategy to use to extract faces from each 
                image. The available options are:

                    * "all" - all faces will be extracted form each 
                      image.
                    * "best" - one face with the largest confidence 
                      score will be extracted from each image.
                    * "largest" - one face with the largest face area 
                      will be extracted from each image.

                For more info, see :meth:`.RetinaFace.__init__`. 
                Defaults to "largest".
            padding: The padding type (border mode) to apply when 
                cropping out faces. If faces are near edge, some part of 
                the resulting center-cropped face image may be blank, in 
                which case it can be padded with specific values. For 
                available options, see 
                `OpenCV BorderTypes <https://docs.opencv.org/3.4/d2/de8/group__core__array.html#ga209f2f4869e304c82d07739337eae7c5>`_. 
                If specified as "constant", the value of 0 will be used. 
                Defaults to "reflect".
            allow_skew: Whether to allow skewing when aligning the face 
                according to its landmarks. If True, then facial points 
                will be matched very closely to the ideal standard 
                landmark points (which is a set of reference points 
                created internally when preforming the transformation). 
                If all faces face forward, i.e., in portrait-like 
                manner, then this could be set to True which results in 
                minimal perspective changes. However, most of the time 
                this should be set to False to preserve the face 
                perspective. For more details, see 
                :meth:`.crop_align`. Defaults to False.
            landmarks: If landmarks are already known, they should be 
                specified via this variable. If specified, landmark 
                estimation will not be performed. There are 2 ways to 
                specify landmarks:

                    1. As a path to landmarks file, in which case str 
                       should be provided. The specified file should 
                       contain file (image) names and corresponding 
                       landmark coordinates. Duplicate file names are 
                       allowed (in case multiple faces are present in 
                       the same image). For instance, it could be 
                       .txt file where each row contains space-separated 
                       values: the first value is the file name and the 
                       other 136 values represent landmark coordinates 
                       in x1, y1, x2, y2, ... format. For more details 
                       about the possible file formats and how they are 
                       parsed, see 
                       :func:`~.utils.parse_landmarks_file`.
                    2. As a tuple of 2 numpy arrays. The first one is of 
                       shape (``num_faces``, ``num_landm``, 2) of type 
                       :attr:`numpy.float32` and represents the 
                       landmarks of every face that is going to be 
                       extracted from images. The second is a numpy 
                       array of shape (``num_faces``,) of type 
                       :class:`numpy.str_` where each value specifies a 
                       file name to which a corresponding set of 
                       landmarks belongs.

                If not specified, 5 landmark coordinates will be 
                estimated for each face automatically. Defaults to None.
            attr_groups: Attribute groups dictionary that specifies how
                to group the output face images according to some common
                attributes. The keys are names describing some common
                attribute, e.g., "glasses", "no_accessories" and the 
                values specify which attribute indices belong (or don't 
                belong, if negative) to that group, e.g., [6], 
                [-6, -9, -15]. For more information, see 
                :class:`.BiSeNet` and :meth:`save_groups`. 
                If not provided, output images will not be grouped by 
                attributes and no attribute sub-folders will be created
                in the desired output directory. Defaults to None.
            mask_groups: Mask groups dictionary that specifies how to 
                group the output face images according to some face
                attributes that make up a segmentation mask. The keys 
                are mask type names, e.g., "eyes", and the values 
                specify which attribute indices should be considered for 
                that mask, e.g., [4, 5]. For every group, not only face 
                images will be saved in a corresponding sub-directory, 
                but also black and white face attribute masks (white 
                pixels indicating the presence of a mask attribute). For 
                more details, see For more info, see 
                :class:`.BiSeNet` and :py:meth:`save_groups`.
                If not provided, no grouping is applied. Defaults to 
                None.
            det_threshold: The visual threshold, i.e., minimum 
                confidence score, for a detected face to be considered 
                an actual face. See :meth:`.RetinaFace.__init__` for 
                more details. If None, no face detection will be 
                performed. Defaults to 0.6.
            enh_threshold: Quality enhancement threshold that tells when 
                the image quality should be enhanced (it is an expensive 
                operation). It is the minimum average face factor, i.e., 
                face area relative to the image, below which the whole 
                image is enhanced. Defaults to 0.001.
            batch_size: The batch size. It is the maximum number of 
                images that can be processed by every processor at a 
                single time-step. Large values may result in memory 
                errors, especially, when GPU acceleration is used. 
                Increase this if less models (i.e., landmark detection, 
                quality enhancement, face parsing models) are used and 
                decrease otherwise. Defaults to 8.
            num_processes: The number of processes to launch to perform 
                image processing. Each process works in parallel on 
                multiple threads, significantly increasing the 
                performance speed. Increase if less prediction models 
                are used and increase otherwise. Defaults to 1.
            device: The device on which to perform the predictions, 
                i.e., landmark detection, quality enhancement and face 
                parsing. If landmarks are provided, no enhancement and 
                no parsing is desired, then this has no effect. Defaults
                to "cpu".
        """
        # Init specified attributes
        self.output_size = output_size
        self.output_format = output_format
        self.resize_size = resize_size
        self.face_factor = face_factor
        self.strategy = strategy
        self.padding = padding
        self.allow_skew = allow_skew
        self.landmarks = landmarks
        self.attr_groups = attr_groups
        self.mask_groups = mask_groups
        self.det_threshold = det_threshold
        self.enh_threshold = enh_threshold
        self.batch_size = batch_size
        self.num_processes = num_processes
        self.device = device

        # The only option for STD
        self.num_std_landmarks = 5

        # Modify attributes to have proper type
        if isinstance(self.output_size, int):
            self.output_size = (self.output_size, self.output_size)
        
        if len(self.output_size) == 1:
            self.output_size = (self.output_size[0], self.output_size[0])
        
        if isinstance(self.resize_size, int):
            self.resize_size = (self.resize_size, self.resize_size)
        
        if len(self.resize_size) == 1:
            self.resize_size = (self.resize_size[0], self.resize_size[0])

        if isinstance(self.device, str):
            self.device = torch.device(device)

        if isinstance(self.landmarks, str):
            self.landmarks = parse_landmarks_file(self.landmarks)

        # Further attributes
        self._init_models()
        self._init_landmarks_target()
    
    def _init_models(self):
        """Initializes detection, enhancement and parsing models.

        The method initializes 3 models:
            1. If ``self.det_threshold`` is provided and no landmarks 
               are known in advance, the detection model is initialized 
               to estimate 5-point landmark coordinates. For more info, 
               see :class:`.RetinaFace`.
            2. If ``self.enh_threshold`` is provided, the quality 
               enhancement model is initialized. For more info, see
               :class:`.RRDBNet`.
            3. If ``self.attr_groups`` or ``self.mask_groups`` is 
               provided, then face parsing model is initialized. For 
               more info, see :class:`.BiSeNet`.

        Note:
            This is a useful initializer function if multiprocessing is 
            used, in which case copies of all the models can be created 
            on separate cores.
        """
        # Init models as None
        self.det_model = None
        self.enh_model = None
        self.par_model = None

        if torch.cuda.is_available() and self.device.index is not None:
            # Helps to prevent CUDA memory errors
            torch.cuda.set_device(self.device.index)
            torch.cuda.empty_cache()

        if self.det_threshold is not None and self.landmarks is None:
            # If detection threshold is set, we will predict landmarks
            self.det_model = RetinaFace(self.strategy, self.det_threshold)
            self.det_model.load(device=self.device)
        
        if self.enh_threshold is not None:
            # If enhancement threshold is set, we might enhance quality
            self.enh_model = RRDBNet(self.enh_threshold)
            self.enh_model.load(device=self.device)
        
        if self.attr_groups is not None or self.mask_groups is not None:
            # If grouping by attributes or masks is set, use parse model
            args = (self.attr_groups, self.mask_groups, self.batch_size)
            self.par_model = BiSeNet(*args)
            self.par_model.load(device=self.device)
    
    def _init_landmarks_target(self):
        """Initializes target landmarks set.

        This method initializes a set of standard landmarks. Standard, 
        or target, landmarks refer to an average set of landmarks with 
        ideal normalized coordinates for each facial point. The source 
        facial points will be rotated, scaled and translated to match 
        the standard landmarks as close as possible.

        Both source (computed separately for each image) and target 
        landmarks must semantically match, e.g., the left eye coordinate 
        in target landmarks also corresponds to  the left eye coordinate 
        in source landmarks.

        There should be a standard landmarks set defined for a desired 
        number of landmarks. Each coordinate in that set is normalized, 
        i.e., x and y values are between 0 and 1. These values are then 
        scaled based on face factor and resized to match the desired 
        output size as defined by ``self.output_size``.

        Note:
            Currently, only 5 standard landmarks are supported.

        Raises:
            ValueError: If the number of standard landmarks is not 
                supported. The number of standard landmarks is 
                ``self.num_std_landmarks``.
        """
        match self.num_std_landmarks:
            case 5:
                # If the number of std landmarks is 5
                std_landmarks = STANDARD_LANDMARKS_5
            case _:
                # Otherwise the number of STD landmarks is not supported
                raise ValueError(f"Unsupported number of standard landmarks "
                                 f"for estimating alignment transform matrix: "
                                 f"{self.num_std_landmarks}.")
        
        # Apply appropriate scaling based on face factor and out size
        std_landmarks[:, 0] *= self.output_size[0] * self.face_factor
        std_landmarks[:, 1] *= self.output_size[1] * self.face_factor

        # Add an offset to standard landmarks to center the cropped face
        std_landmarks[:, 0] += (1 - self.face_factor) * self.output_size[0] / 2
        std_landmarks[:, 1] += (1 - self.face_factor) * self.output_size[1] / 2

        # Pass STD landmarks as target landms
        self.landmarks_target = std_landmarks

    def crop_align(
        self,
        images: np.ndarray | list[np.ndarray],
        padding: np.ndarray | None,
        indices: list[int],
        landmarks_source: np.ndarray,
    ) -> np.ndarray:
        """Aligns and center-crops faces based on the given landmarks.

        This method takes a batch of images (can be padded), and loops 
        through each image (represented as a numpy array) performing the 
        following actions:

            1. Removes the padding.
            2. Estimates affine transformation from source landmarks to 
               standard landmarks.
            3. Applies transformation to align and center-crop the face 
               based on the face factor.
            4. Returns a batch of face images represented as numpy 
               arrays of the same length ans the number of sets of 
               landmarks.

        Crucial role in this method plays ``self.landmarks_target`` 
        which is the standard set of landmarks used as a reference for 
        the source landmarks. Target and source landmark sets are used 
        to estimate transformations of images - each image to which a 
        set of landmarks (from source landmarks batch) belongs is 
        transformed such that the area covers the those landmarks as the 
        standard  (target) landmarks set (as ideally as possible). For 
        more details about target landmarks, check 
        :meth:`_init_landmarks_target`.

        Note:
            If ``self.allow_skew`` is set to True, then facial points 
            will also be skewed to match ``self.landmarks_target`` as 
            close as possible (resulting in, e.g., longer/flatter faces 
            than in the original images).

        Args:
            images: Image batch of shape (N, H, W, 3) of type 
                :attr:`numpy.uint8` (doesn't matter if RGB or BGR) where 
                each nth image is transformed to extract face(-s). 
                (H, W) should be ``self.resize_size``. It can also be a 
                list of :attr:`numpy.uint8` numpy arrays of different 
                shapes.
            padding: Padding of shape (N, 4) where the integer values 
                correspond to the number of pixels padded from each 
                side: top, bottom, left, right. Padding was originally 
                applied to each image, e.g., to make the image square, 
                so that all images could be stacked as a batch. 
                Therefore, it is needed here to remove the padding. If 
                specified as None, it is assumed that the images are 
                un-padded.
            indices: Indices list of length num_faces where each index 
                specifies which image is used to extract faces for each 
                set of landmarks in ``landmarks_source``.
            landmarks_source: Landmarks batch of shape 
                (num_faces, ``self.num_std_landmarks``, 2). These are 
                landmark sets of all the desired faces to extract from 
                the given batch of N images.

        Returns:
            A batch of aligned and center-cropped faces where the factor 
            of the area of a face relative to the whole face image area
            is ``self.face_factor``. The output is a numpy array of 
            shape (N, H, W) of type :attr:`numpy.uint8` (same channel 
            structure as for the input images). (H, W) is defined by 
            ``self.output_size``.
        """
        # Init list, border mode
        transformed_images = []
        border_mode = getattr(cv2, f"BORDER_{self.padding.upper()}")
        
        for landmarks_idx, image_idx in enumerate(indices):
            if self.allow_skew:
                # Perform full perspective transformation
                transform_function = cv2.estimateAffine2D
            else:
                # Preform only rotation, scaling and translation
                transform_function = cv2.estimateAffinePartial2D
            
            # Estimate transformation matrix to apply
            transform_matrix = transform_function(
                landmarks_source[landmarks_idx],
                self.landmarks_target,
                ransacReprojThreshold=np.inf,
            )[0]

            if transform_matrix is None:
                # Could not estimate
                continue

            # Retrieve current image
            image = images[image_idx]

            if padding is not None:
                # Crop out the un-padded area
                [t, b, l, r] = padding[image_idx]
                image = image[t:image.shape[0]-b, l:image.shape[1]-r]

            # Apply affine transformation to the image
            transformed_images.append(cv2.warpAffine(
                image,
                transform_matrix,
                self.output_size,
                borderMode=border_mode
            ))
        
        # Normally stacking would be applied unless the list is empty
        numpy_fn = np.stack if len(transformed_images) > 0 else np.array
        
        return numpy_fn(transformed_images)
    
    def save_group(
        self,
        faces: np.ndarray,
        file_names: list[str],
        output_dir: str,
    ):
        """Saves a group of images to output directory.

        Takes in a batch of faces or masks as well as corresponding file 
        names from where the faces were extracted and saves the 
        faces/masks to a specified output directory with the same names 
        as those image files (appends counter suffixes if multiple faces 
        come from the same file). If the batch of face images/masks is 
        empty, then the output directory is not created either.

        Args:
            faces: Face images (cropped and aligned) represented as a
                numpy array of shape (N, H, W, 3) with values of type
                :attr:`numpy.uint8` ranging from 0 to 255. It may also 
                be face mask of shape (N, H, W) with values of 255 where 
                some face attribute is present and 0 elsewhere.
            file_names: The list of filenames of length N. Each face 
                comes from a specific file whose name is also used to 
                save the extracted face. If ``self.strategy`` allows 
                multiple faces to be extracted from the same file, such 
                as "all", counters at the end of filenames are added.
            output_dir: The output directory to save ``faces``.
        """
        if len(faces) == 0:
            # Just return
            return
        
        # Create output directory, name counts
        os.makedirs(output_dir, exist_ok=True)
        file_name_counts = defaultdict(lambda: -1)

        for face, file_name in zip(faces, file_names):
            # Split each filename to base name, ext
            name, ext = os.path.splitext(file_name)

            if self.output_format is not None:
                # If specific img format given
                ext = '.' + self.output_format

            if self.strategy == "all":
                # Attach numbering to filenames
                file_name_counts[file_name] += 1
                name += f"_{file_name_counts[file_name]}"
            
            if face.ndim == 3:
                # If it's a colored img (not a mask), to BGR
                face = cv2.cvtColor(face, cv2.COLOR_RGB2BGR)

            # Make image path based on file format and save
            file_path = os.path.join(output_dir, name + ext)
            cv2.imwrite(file_path, face)
    
    def save_groups(
        self,
        faces: np.ndarray,
        file_names: np.ndarray,
        output_dir: str,
        attr_groups: dict[str, list[int]] | None,
        mask_groups: dict[str, tuple[list[int], np.ndarray]] | None,
    ):
        """Saves images (and masks) group-wise.

        This method takes a batch of face images of equal dimensions, a 
        batch of file names identifying which image each face comes 
        from, and, optionally, attribute and/or mask groups telling how 
        to split the face images (and masks) across different folders.
        This method then loops through all the groups and saves images 
        accordingly.

        Example 1:
            If neither ``attr_groups`` nor ``mask_groups`` are provided, 
            the face images will be saved according to this structure::

                ├── output_dir
                |    ├── face_image_0.jpg
                |    ├── face_image_1.png
                |    ...

        Example 2:
            If only ``attr_groups`` is provided (keys are names 
            describing common attributes across faces in that group and 
            they are also sub-directories of ``output_dir``), the 
            structure is as follows::

                ├── output_dir
                |    ├── attribute_group_1
                |    |    ├── face_image_0.jpg
                |    |    ├── face_image_1.png
                |    |    ...
                |    ├── attribute_group_2
                |    ...

        Example 3:
            If only ``mask_groups`` is provided (keys are names 
            describing the mask type and they are also sub-directories 
            of ``output_dir``), the structure is as follows::

                ├── output_dir
                |    ├── group_1
                |    |    ├── face_image_0.jpg
                |    |    ├── face_image_1.png
                |    |    ...
                |    ├── group_1_mask
                |    |    ├── face_image_0.jpg
                |    |    ├── face_image_1.png
                |    |    ...
                |    ├── group_2
                |    |    ...
                |    ├── group_2_mask
                |    |    ...
                |    ...

        Example 4:
            If both ``attr_groups`` and ``mask_groups`` are provided, 
            then all images and masks will first be grouped by 
            attributes and then by mask groups. The structure is then as 
            follows::

                ├── output_dir
                |    ├── attribute_group_1
                |    |    ├── group_1_mask
                |    |    |    ├── face_image_0.jpg
                |    |    |    ├── face_image_1.png
                |    |    |    ...
                |    |    ├── group_1_mask
                |    |    |    ├── face_image_0.jpg
                |    |    |    ├── face_image_1.png
                |    |    |    ...
                |    |    ├── group_2
                |    |    |    ...
                |    |    ├── group_2_mask
                |    |    |    ...
                |    |    ...
                |    |
                |    ├── attribute_group_2
                |    |    ...
                |    ...

        Args:
            faces: Face images (cropped and aligned) represented as a
                numpy array of shape (N, H, W, 3) with values of type
                :attr:`numpy.uint8` ranging from 0 to 255.
            file_names: File names of images from which the faces were 
                extracted. This value is a numpy array of shape (N,) 
                with values of type :class:`numpy.str_`. Each nth 
                face in ``faces`` maps to exactly one file nth name in
                this array, thus there may be duplicate file names
                (because different faces may come from the same file).
            output_dir: The output directory where the faces or folders 
                of faces will be saved to.
            attr_groups: Face groups by attributes. Each key represents 
                the group name (describes common attributes across
                faces) and each value is a list of indices identifying 
                faces (from `faces`) that should go to that group.
            mask_groups: Face groups by extracted masks. Each key
                represents group name (describes the mask type) and each 
                value is a tuple where the first element is a list of 
                indices identifying faces (from ``faces``) that should 
                go to that group and the second element is a batch of 
                masks corresponding to indexed faces represented as 
                numpy arrays of shape (N, H, W) with values of type 
                :attr:`numpy.uint8` and being either 0 (negative) or 255 
                (positive).
        """
        if attr_groups is None:
            # No-name group of idx mapping to all faces
            attr_groups = {'': list(range(len(faces)))}
        
        if mask_groups is None:
            # No-name group mapping to all faces, with no masks
            mask_groups = {'': (list(range(len(faces))), None)}

        for attr_name, attr_indices in attr_groups.items():
            for mask_name, (mask_indices, masks) in mask_groups.items():
                # Make mask group values that fall under attribute group
                group_idx = list(set(attr_indices) & set(mask_indices))
                group_dir = os.path.join(output_dir, attr_name, mask_name)

                # Retrieve group values & save
                face_group = [faces[idx] for idx in group_idx]
                file_name_group = file_names[group_idx]
                self.save_group(face_group, file_name_group, group_dir)

                if masks is not None:
                    # Save to masks dir
                    group_dir += "_mask"
                    masks = masks[[mask_indices.index(i) for i in group_idx]]
                    self.save_group(masks, file_name_group, group_dir)
    
    def process_batch(self, file_names: list[str], input_dir: str, output_dir: str):
        """Extracts faces from a batch of images and saves them.

        Takes file names, input directory, reads images and extracts 
        faces and saves them to the output directory. This method works 
        as follows:

            1. *Batch generation* - a batch of images form the given 
               file names is generated. Each images is padded and 
               resized to ``self.resize_size`` while keeping the same 
               aspect ratio.
            2. *Landmark detection* - detection model is used to predict 
               5 landmarks for each face in each image, unless the 
               landmarks were already initialized  or face alignment + 
               cropping is not needed.
            3. *Image enhancement* - some images are enhanced if the 
               faces compared with the image size are small. If 
               landmarks are None, i.e., if no alignment + cropping was 
               desired, all images are enhanced. Enhancement is not done 
               if ``self.enh_threshold`` is None.
            4. *Image grouping* - each face image is parsed, i.e., a map 
               of face attributes is generated. Based on those 
               attributes, each face image is put to a corresponding 
               group. There may also be mask groups, in which case masks 
               for each image in that group are also generated. Faces 
               are not parsed if ``self.attr_groups`` and 
               ``self.mask_groups`` are both None.
            5. *Image saving* - each face image (and a potential mask) 
               is saved according to the group structure (if there is 
               any).
        
        Note:
            If detection model is not used, then batch is just a list of 
            loaded images of different dimensions.

        Args:
            file_names: The list of image file names (not full paths). 
                All the images should be in the same directory.
            input_dir: Path to input directory with image files.
            output_dir: Path to output directory to save the extracted 
                face images.
        """
        # Read images and filter valid corresponding file names
        images, file_names = read_images(file_names, input_dir)

        if self.landmarks is None and self.det_model is None:
            # One-to-one image to index mapping and no landmarks
            indices, landmarks = list(range(len(file_names))), None
        elif self.landmarks is not None:
            # Initialize empty idx lists
            indices, indices_ldm = [], []

            for i, file_name in enumerate(file_names):
                # Check the indices of landmark sets in landmarks file
                indices_i = np.where(file_name == self.landmarks[1])[0]

                # Update img & ldm file name indices
                indices.extend([i] * len(indices_i))
                indices_ldm.extend(indices_i.tolist())
            
            # Set landmarks according to the indices
            landmarks = self.landmarks[0][indices_ldm]
            
        elif self.det_model is not None:
             # Create a batch of images (with faces) and their paddings
            images, _, paddings = as_batch(images, self.resize_size)
            images, paddings = as_tensor(images, self.device), paddings

            # If landmarks were not given, predict, undo padding
            landmarks, indices = self.det_model.predict(images)
            landmarks -= paddings[indices][:, None, [2, 0]]

        if landmarks is not None and len(landmarks) == 0:
            # Nothing to save
            return
            
        if landmarks is not None and landmarks.shape[1] != self.num_std_landmarks:
            # Compute the mean landmark coordinates from retrieved slices
            slices = get_ldm_slices(self.num_std_landmarks, landmarks.shape[1])            
            landmarks = np.stack([landmarks[:, s].mean(1) for s in slices], 1)

        if self.enh_model is not None:
            # Enhance some images
            images = as_tensor(images, self.device)
            images = self.enh_model.predict(images, landmarks, indices)

        # Convert to numpy images, initialize groups
        images, groups = as_numpy(images), (None, None)

        if landmarks is not None:    
            # Generate source, target landmarks, estimate & apply transform
            images = self.crop_align(images, paddings, indices, landmarks)

        if self.par_model is not None:
            # Predict attribute and mask groups if face parsing desired
            groups = self.par_model.predict(as_tensor(images, self.device))

        # Pick file names for each face, save faces (by groups if exist)
        self.save_groups(images, file_names[indices], output_dir, *groups)
    
    def process_dir(self, input_dir: str, output_dir: str | None = None):
        """Processes images in the specified input directory.

        Splits all the file names in the input directory to batches 
        and processes batches on multiple cores. For every file name 
        batch, images are loaded, some are optionally enhanced, 
        landmarks are generated and used to optionally align and 
        center-crop faces, and grouping is optionally applied based on
        face attributes. For more details, check 
        :meth:`process_batch`.

        Note:
            There might be a few seconds delay before the actual 
            processing starts if there are a lot of files in the 
            directory - it takes some time to split all the file names 
            to batches.

        Args:
            input_dir: Path to input directory with image files.
            output_dir: Path to output directory to save the extracted 
                (and optionally grouped to sub-directories) face images. 
                If None, then the same path as for ``input_dir`` is used 
                and additionally "_faces" suffix is added to the name.
        """
        if output_dir is None:
            # Create a default output dir name
            output_dir = input_dir + "_faces"

        # Create batches of image file names in input dir
        files, bs = os.listdir(input_dir), self.batch_size
        file_batches = [files[i:i+bs] for i in range(0, len(files), bs)]

        if len(file_batches) == 0:
            # Empty
            return
        
        # Define worker function and its additional arguments
        kwargs = {"input_dir": input_dir, "output_dir": output_dir}
        worker = partial(self.process_batch, **kwargs)
        
        with ThreadPool(self.num_processes, self._init_models) as pool:
            # Create imap object and apply workers to it
            imap = pool.imap_unordered(worker, file_batches)
            list(tqdm.tqdm(imap, total=len(file_batches), desc="Processing"))



import torch
import numpy as np
import torch.nn as nn
import torch.nn.functional as F
from ._layers import LoadMixin, RRDB


class RRDBNet(nn.Module, LoadMixin):
    """Face quality enhancer.

    This model is capable of detecting which images have low-quality 
    faces, i.e., which images have small face areas compared to the 
    dimensions of the image and is able to enhance the quality of those 
    images. The images are up-scaled 4 times and then resized to their 
    original size - this results in less blurry faces.

    This class also inherits ``load`` method from ``LoadMixin`` class. 
    The method takes a device on which to load the model and loads the 
    model with a default state dictionary loaded from 
    ``WEIGHTS_FILENAME`` file. It sets this model to eval mode and 
    disables gradients.

    For more information on how RetinaFace model works, see this repo:
    `BSRGAN <https://github.com/cszn/BSRGAN>`_. Most of the code was 
    taken from that repository.
    
    Note:
        Whenever an input shape is mentioned, N corresponds to batch 
        size, C corresponds to the number of channels, H - to input
        height, and W - to input width.
    """
    #: WEIGHTS_FILENAME (str): The constant specifying the name of 
    #: ``.pth`` file from which the weights for this model should be 
    #: loaded. Defaults to "bsrgan_x4_enhancer.pth".
    WEIGHTS_FILENAME = "bsrgan_x4_enhancer.pth"

    def __init__(self, min_face_factor: float = 0.001):
        """Initializes RRDB (BSRGAN) model.

        Just assigns the minimum face threshold attribute and constructs 
        module layers for quality inference.

        Args:
            min_face_factor: The minimum average face factor, i.e., face 
                area relative to the image, below which the whole image 
                is enhanced. Defaults to 0.001.
        """
        super().__init__()
        # Init minimum face factor attribute
        self.min_face_factor = min_face_factor

        # Initialize first layers that produce features
        self.conv_first = nn.Conv2d(3, 64, 3, 1, 1)
        self.RRDB_trunk = nn.Sequential(*[RRDB(64, 32) for _ in range(23)])
        self.trunk_conv = nn.Conv2d(64, 64, 3, 1, 1)
        self.upconv1 = nn.Conv2d(64, 64, 3, 1, 1)
        self.upconv2 = nn.Conv2d(64, 64, 3, 1, 1)

        # Final layers that produce enhanced image
        self.HRconv = nn.Conv2d(64, 64, 3, 1, 1)
        self.conv_last = nn.Conv2d(64, 3, 3, 1, 1)
        self.lrelu = nn.LeakyReLU(negative_slope=0.2, inplace=True)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Performs forward pass.

        Takes an input tensor which is a batch of images and produces 
        the same batch, except images are up-scaled 4 times. 

        Args:
            x: The input tensor of shape (N, 3, H, W).

        Returns:
            An output tensor of shape (N, 3, 4*H, 4*W).
        """
        # Perform inference, get features, upscale 2 times, get enhanced
        fea = (x := self.conv_first(x)) + self.trunk_conv(self.RRDB_trunk(x))
        fea = self.lrelu(self.upconv1(F.interpolate(fea, scale_factor=2)))
        fea = self.lrelu(self.upconv2(F.interpolate(fea, scale_factor=2)))

        return self.conv_last(self.lrelu(self.HRconv(fea)))

    @torch.no_grad()
    def predict(
        self,
        images: torch.Tensor | list[torch.Tensor],
        landmarks: np.ndarray | None,
        indices: list[int] | None,
    ) -> torch.Tensor:
        """Enhances the quality of images with low-quality faces.

        Takes a batch of images and sets of landmarks for each image and 
        enhances the quality of those images for which the average face 
        area factor is lower than ``self.min_face_factor``. The face 
        factor is computed by dividing the face area (computed by 
        multiplying the width and the height of the face, specified by 
        left-eye, right-eye, left-mouth, right-mouth landmark 
        coordinates) by the image area.

        Note:
            The images are enhanced one by one instead of as a batch 
            because the inference is very memory consuming and can 
            result in memory errors.

        Args:
            images: Image batch of shape (N, 3, H, W) in RGB form with 
                float values from 0.0 to 255.0. It must be on the same 
                device as this model. It can also be a list of tensors
                of different shapes.
            landmarks: Landmarks batch of shape (``num_faces``, 5, 2) 
                used to compute average face area for each image. If 
                None, then every image will be enhanced.
            indices: Indices list mapping each set of landmarks to a 
                specific image in ``images`` batch (multiple sets of 
                landmarks can come from the same image). If None, then 
                every image will be enhanced.

        Returns:
            The same image batch as ``images`` - the shape is 
            (N, 3, H, W) channels are in RGB and values range from 
            0.0 to 255.0. The only difference is that some of the images 
            are of much higher quality, i.e., less blurry.
        """
        for i in range(len(images)):
            if landmarks is None or indices is None:
                # Create a dummy face factor to ensure enhance
                face_factor = np.array([self.min_face_factor])
            else:
                # Select all face landmarks in the current i'th image
                landmarks_i = landmarks[[idx == i for idx in indices]]

                if len(landmarks_i) == 0:
                    # No landmarks found
                    continue

                # Compute relative face factor (area face takes up)
                [w, h] = (landmarks_i[:, 4] - landmarks_i[:, 0]).T
                face_factor = w * h / (images[0].shape[1] * images[0].shape[2])

            if face_factor.mean() <= self.min_face_factor:
                # Enhance ith img if factor below threshold
                image_x4 = self(images[i].unsqueeze(0).div(255))
                image_x1 = F.interpolate(image_x4, None, 0.25, "bicubic")
                images[i] = image_x1.clamp(0, 1).mul(255).round()[0]

        return images



import torch
import numpy as np
import torch.nn as nn
import torch.nn.functional as F
from ._layers import LoadMixin, ContextPath, FeatureFusionModule, BiSeNetOutput


class BiSeNet(nn.Module, LoadMixin):
    """Face attribute parser.

    This class is capable of predicting scores for 19 attributes for 
    face images. After it identifies the closest attribute for each 
    pixel, it can also put the whole face image to a corresponding 
    attribute or mask group.

    The 19 attributes are as follows (attributes are indicated from a  
    person's face perspective, meaning, for instance, left eye is the
    eye on the right hand-side of the picture, however, sides are not 
    always accurate):

        * 0 - neutral
        * 1 - skin
        * 2 - left eyebrow
        * 3 - right eyebrow
        * 4 - left eye
        * 5 - right eye
        * 6 - eyeglasses
        * 7 - left ear
        * 8 - right ear
        * 9 - earing
        * 10 - nose
        * 11 - mouth
        * 12 - upper lip
        * 13 - lower lip
        * 14 - neck
        * 15 - neckless
        * 16 - clothes
        * 17 - hair
        * 18 - hat

    Some examples of grouping by attributes:

        * ``'glasses': [6]`` - this will put each face image that 
          contains pixels labeled as 6 to a category called 'glasses'.
        * ``'earings_and_neckless': [9, 15]`` - this will put each image 
          that contains pixels labeled as 9 and also contains pixels
          labeled as 15 to a category called 'earings_and_neckless'.
        * ``'no_accessories': [-6, -9, -15, -18]`` - this will put each 
          face image that does not contain pixels labeled as either 6, 
          9, 15, or 18 to a category called 'no_accessories'.
    
    Some examples of grouping by mask:

        * ``'nose': [10]`` - this will put each face image that contains 
          pixels labeled as 10 to a category called 'nose' and generate 
          a corresponding mask.  
        * ``'eyes_and_eyebrows': [2, 3, 4, 5]`` - this will put each 
          image that contains pixels labeled as either 2, 3, 4, or 5 (or 
          any combination of them) to a category called 
          'eyes_and_eyebrows' and generate a corresponding mask.
    
    This class also inherits ``load`` method from ``LoadMixin`` class. 
    The method takes a device on which to load the model and loads the 
    model with a default state dictionary loaded from 
    ``WEIGHTS_FILENAME`` file. It sets this model to eval mode and 
    disables gradients.
    
    For more information on how BiSeNet model works, see this repo:
    `Face Parsing PyTorch <https://github.com/zllrunning/face-parsing.PyTorch>`_. 
    Most of the code was taken from that repository.

    Note:
        Whenever an input shape is mentioned, N corresponds to batch 
        size, C corresponds to the number of channels, H - to input
        height, and W - to input width.

    Be default, this class initializes the following attributes which 
    can be changed after initialization of the class (but, typically, 
    should not be changed):
    
    Attributes:
        attr_join_by_and (bool): Whether to add a face image to 
            an attribute group if the face meets all the specified 
            attributes in a list (joined by and) of at least one of 
            the attributes (joined by or). Please read the definition 
            of  `attr_groups` to get a clearer picture. In most cases, 
            this should be set True - if the attributes in a group 
            list are negative, this will ensure the selected face will 
            match none of the specified attributes. Also, if you want 
            to join the attributes by or (any), then separate 
            single-attribute groups can be created and manually merged 
            into one. Defaults to True.
        attr_threshold (int): Threshold, based on which the 
            attribute is determined as present in the face image. For 
            instance, if the threshold is 5, then at least 6 pixels 
            must be labeled of the same kind of attribute for that 
            attribute to be considered present in the face image. 
            Defaults to 5.
        mask_threshold (int): Threshold, based on which the 
            mask is considered to be a proper mask. For instance, if 
            the threshold is 15, then face images for which the number 
            of pixels with the values corresponding to a specified 
            mask group (face attributes) is less than or equal to 15 
            will be ignored and image-mask pair for that mask category 
            will not be generated. Defaults to 15.
        mean (list[float]): The list of mean values for each 
            input channel. The pixel values should be shifted by those 
            quantities during inference since this normalization was 
            applied during training. Defaults to 
            [0.485, 0.456, 0.406].
        std (list[float]): The list of standard deviation values 
            for each input channel. The pixel values should be scaled 
            by those quantities during inference since this 
            normalization was applied during training. Defaults to 
            [0.229, 0.224, 0.225].
    """
    #: WEIGHTS_FILENAME (str): The constant specifying the name of 
    #: ``.pth`` file from which the weights for this model should be 
    #: loaded. Defaults to "bise_parser.pth".
    WEIGHTS_FILENAME = "bise_parser.pth"

    def __init__(
        self,
        attr_groups: dict[str, list[int]] | None = None,
        mask_groups: dict[str, list[int]] | None = None,
        max_batch_size: int = 8,
    ):
        """Initializes BiSeNet model.

        First it assigns the passed values as attributes. Then this 
        method initializes BiSeNet layers required for face parsing, 
        i.e., labeling face parts.

        Note:
            Check class definition for the possible face attribute 
            values and examples of groups. Also note that all the 
            specified variables here are mainly relevant only for 
            :meth:`predict`.

        Args:
            attr_groups: Dictionary specifying attribute groups, based 
                on which the face images should be grouped. Each key 
                represents an attribute group name, e.g., 'glasses', 
                'earings_and_neckless', 'no_accessories', and each value 
                represents attribute indices, e.g., `[6]`, `[9, 15]`, 
                `[-6, -9, -15, -18]`, each index mapping to some 
                attribute. Since this model labels face image pixels, if 
                there is enough pixels with the specified values in the 
                list, the whole face image will be put into that 
                attribute category. For negative values, it will be 
                checked that the labeled face image does not contain
                those (absolute) values. If it is None, then there will 
                be no grouping according to attributes. Defaults to
                None.
            mask_groups: Dictionary specifying mask groups, based on 
                which the face images and their masks should be grouped. 
                Each key represents a mask group name, e.g., 'nose', 
                'eyes_and_eyebrows', and each value represents attribute 
                indices, e.g., `[10]`, `[2, 3, 4, 5]`, each index
                mapping to some attribute. Since this model labels face 
                image pixels, a mask will be created with ones at pixels 
                that match the specified attributes and zeros elsewhere.
                Note that negative values would make no sense here and 
                having them would cause an error. If it is None, then 
                there will be no grouping according to mask groups. 
                Defaults to None.
            max_batch_size: The maximum batch size used when performing 
                inference. There may be a lot of faces, in a single 
                batch thus splitting to sub-batches for inference and 
                then merging back predictions is a way to deal with 
                memory errors. This is a convenience variable because 
                batch size typically corresponds to the number of images 
                for a single inference, but the input given in 
                :meth:`predict` might have a larger batch 
                size because it represents the number of faces, many of 
                which can come from just a single image. Defaults to 8.
        """
        super().__init__()
        
        # Initialize class attributes
        self.attr_groups = attr_groups
        self.mask_groups = mask_groups
        self.batch_size = max_batch_size
        self.attr_join_by_and = True
        self.attr_threshold = 5
        self.mask_threshold = 10
        self.mean = [0.485, 0.456, 0.406]
        self.std = [0.229, 0.224, 0.225]

        # Init model layers
        self.cp = ContextPath()
        self.ffm = FeatureFusionModule(256, 256)
        self.conv_out = BiSeNetOutput(256, 256, 19)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Performs forward pass.

        Takes an input batch and performs inference based on the modules 
        it has. The input is a batch of face images and the output is a 
        corresponding batch of pixel-wise attribute scores. 

        Args:
            x: The input tensor of shape (N, 3, H, W).

        Returns:
            An output tensor of shape (N, 19, H, W) where each channel 
            corresponds to a specific attribute and each value at 
            (H, W) is an unbounded confidence score.
        """
        # Generate final features from layers, upscale
        feat_out = self.conv_out(self.ffm(*self.cp(x)))
        return F.interpolate(feat_out, x.size()[2:], None, "bilinear", True)
    
    def group_by_attributes(
        self,
        parse_preds: torch.Tensor,
        attr_groups: dict[str, list[int]],
        offset: int,
    ) -> dict[str, list[int]]:
        """Groups parse predictions by face attributes.

        Takes parse predictions for each face where each pixel 
        corresponds to some attribute group (the integer value 
        indicates that group) and extends the groups in attribute 
        dictionary to include more samples that match the group.

        Args:
            parse_preds: Face parsing predictions of shape (N, H, W) 
                with integer values indicating pixel categories.
            attr_groups: The dictionary with keys corresponding to 
                attribute group names (they match ``self.attr_groups`` 
                keys) and values corresponding to indices that map face
                images from other batches of ``parse_preds`` to the
                corresponding group. This is the dictionary that is 
                extended and returned.
            offset: The offset to add to each index. Originally, the
                indices will correspond only to the face parsings in the 
                current ``parse_preds`` batch and the offset allows to 
                generalize the each index by offsetting it by the 
                previous number of processes face parsings, i.e., the 
                offset is the number of previous batches 
                (``parse_preds``) times the batch size.
        Returns:
            Similar to ``attr_groups``, it is the dictionary with the 
            same keys but values (which are lists of indices) may be 
            extended with additional indices.
        """
        # Specify function/criteria to join the attributes in a list
        att_join = torch.all if self.attr_join_by_and else torch.any
        
        for k, v in self.attr_groups.items():
            # Get the list of face attributes per group and count pixels
            attr = torch.tensor(v, device=parse_preds.device).view(1, -1, 1, 1)
            is_attr = (parse_preds.unsqueeze(1) == attr.abs()).sum(dim=(2, 3))

            # Compare each face against each attribute in a  group
            is_attr = att_join(torch.stack([
                is_attr[:, i] > self.attr_threshold if a > 0 else
                is_attr[:, i] <= self.attr_threshold
                for i, a in enumerate(v)
            ], dim=1), dim=1)

            # Add indices of those faces which match the group attribute
            inds = [i + offset for i in range(len(is_attr)) if is_attr[i]]
            attr_groups[k].extend(inds)
        
        return attr_groups

    def group_by_masks(
        self,
        parse_preds: torch.Tensor,
        mask_groups: dict[str, tuple[list[int], list[np.ndarray]]],
        offset: int,
    ) -> dict[str, tuple[list[int], list[np.ndarray]]]:
        """Groups parse predictions by face mask attributes.

        Takes parse predictions for each face where each pixel 
        corresponds to some parse/mask group (the integer value 
        indicates that group) and extends the groups in mask 
        dictionary to include more samples that match the group.

        Args:
            parse_preds: Face parsing predictions of shape (N, H, W) 
                with integer values indicating pixel categories.
            mask_groups: The dictionary with keys corresponding to 
                mask group names (they match ``self.mask_groups`` keys)
                and values corresponding to tuples where the first value 
                is a list of indices that map face images from other 
                batches of ``parse_preds`` to the corresponding group 
                and the second is a list of corresponding masks as numpy 
                arrays of shape (H, W) of type :attr:`numpy.uint8` with 
                255 at pixels that match the mask group specification 
                and 0 elsewhere. This is the dictionary that is extended 
                and returned.
            offset: The offset to add to each index. Originally, the
                indices will correspond only to the face parsings in the 
                current ``parse_preds`` batch and the offset allows to 
                generalize the each index by offsetting it by the 
                previous number of processes face parsings, i.e., the 
                offset is the number of previous batches 
                (``parse_preds``) times the batch size.

        Returns:
            Similar to ``mask_groups``, it is the dictionary with the 
            same keys but values (which are tuples of a list of indices 
            and a list of masks) may be extended with additional indices 
            and masks.
        """
        # Retrieve threshold (shorter name)
        threshold = self.mask_threshold

        for k, v in self.mask_groups.items():
            # Get the list of face attributes per group and check match
            attr = torch.tensor(v, device=parse_preds.device).view(1, -1, 1, 1)
            mask = (parse_preds.unsqueeze(1) == attr).any(dim=1)

            # Identify proper masks and convert them to numpy image type
            inds = [i for i in range(len(mask)) if mask[i].sum() > threshold]
            masks = mask[inds].mul(255).cpu().numpy().astype(np.uint8)

            # Extend the lists of indices and masks for k group
            mask_groups[k][0].extend([i + offset for i in inds])
            mask_groups[k][1].extend([*masks])

        return mask_groups
    
    @torch.no_grad()
    def predict(
        self,
        images: torch.Tensor | list[torch.Tensor],
    ) -> tuple[dict[str, list[int]] | None, 
               dict[str, tuple[list[int], list[np.ndarray]]] | None]:
        """Predicts attribute and mask groups for face images.

        This method takes a batch of face images groups them according 
        to the specifications in ``self.attr_groups`` and 
        ``self.mask_groups``. For more information on how it works, see 
        this class' specification :class:`BiSeNet`. It returns 2 
        groups maps - one for grouping face images to different
        attribute categories, e.g., 'with glasses', 'no accessories' and 
        the other for grouping images to different mask groups, e.g., 
        'nose', 'lips and mouth'.

        Args:
            images: Image batch of shape (N, 3, H, W) in RGB form with 
                float values from 0.0 to 255.0. It must be on the same 
                device as this model. A list of tensors can also be 
                provided, however, they all must have the same spatial 
                dimensions to be stack-able to a single batch.

        Returns:
            A tuple of 2 dictionaries (either can be None):

                1. ``attr_groups`` - each key represents attribute 
                   category and each value is a list of indices 
                   indicating which  samples from ``images`` batch 
                   belong to that category. It can be None if 
                   ``self.attr_groups`` is None.
                2. `mask_groups` - each key represents attribute (mask) 
                   category and each value is a tuple where the first 
                   element is a list of indices indicating which samples 
                   from ``images`` batch belong to that mask group and 
                   the second element is a corresponding batch of masks 
                   of shape (N, H, W) of type :attr:`numpy.uint8` with 
                   values of either 0 or 255. The masks are presented in 
                   that order as the indices indicate which face images 
                   to  take for that mask group. It can be None if 
                   ``self.mask_groups`` is None.

        """
        # Initialize groups as None, a helper offset
        attr_groups, mask_groups, offset = None, None, 0

        if self.attr_groups is not None:
            # Initialize an empty dictionary of attribute groups
            attr_groups = {k: [] for k in self.attr_groups.keys()}
        
        if self.mask_groups is not None:
            # Initialize an empty dictionary of mask groups
            mask_groups = {k: ([], []) for k in self.mask_groups.keys()}
        
        if isinstance(images, list):
            # Stack the list of tensors
            images = torch.stack(images)
        
        # Convert mean and std to tensors and reshape, resize images
        mean = torch.tensor(self.mean, device=images.device).view(1, 3, 1, 1)
        std = torch.tensor(self.std, device=images.device).view(1, 3, 1, 1)
        x = F.interpolate(images.div(255), (512, 512), mode="bilinear")

        for sub_x in torch.split(x, self.batch_size):
            # Inference and resize back
            o = self((sub_x - mean) / std)
            o = F.interpolate(o, images.size()[2:], mode="nearest").argmax(1)

            if self.attr_groups is not None:
                # Extend each attribute group based on predictions
                attr_groups = self.group_by_attributes(o, attr_groups, offset)
            
            if self.mask_groups is not None:
                # Extend each mask group based on predictions
                mask_groups = self.group_by_masks(o, mask_groups, offset)
            
            # Increment offset
            offset += len(sub_x)
        
        if attr_groups is not None:
            # Filter out groups for which the list of indices is empty
            attr_groups = {k: v for k, v in attr_groups.items() if len(v) > 0}
        
        if mask_groups is not None:
            # Filter out groups for which the list of indices is empty
            mask_groups = {
                k: (v[0], np.stack(v[1]))
                for k, v in mask_groups.items() if len(v[1]) > 0
            }

        return attr_groups, mask_groups



import torch
import numpy as np
import torch.nn as nn
import torch.nn.functional as F
import torchvision.models as models
import torchvision.models._utils as _utils
from ._layers import LoadMixin, PriorBox, SSH, FPN, Head


class RetinaFace(nn.Module, LoadMixin):
    """RetinaFace face detector and 5-point landmark predictor.

    This class is capable of predicting 5-point landmarks from a batch 
    of images and filter them based on strategy, e.g., "all landmarks in 
    the image", "a single set of landmarks per image of the largest
    face". For more information, see the main method of this class
    :meth:`predict`. For main attributes, see :meth:`__init__`.

    This class also inherits ``load`` method from ``LoadMixin`` class. 
    The method takes a device on which to load the model and loads the 
    model with a default state dictionary loaded from 
    ``WEIGHTS_FILENAME`` file. It sets this model to eval mode and 
    disables gradients.

    For more information on how RetinaFace model works, see this repo:
    `PyTorch Retina Face <https://github.com/biubug6/Pytorch_Retinaface>`_. 
    Most of the code was taken from that repository.

    Note:
        Whenever an input shape is mentioned, N corresponds to batch 
        size, C corresponds to the number of channels, H - to input
        height, and W - to input width. ``out_dim`` corresponds to the
        total guesses (the number of priors) the model made about each
        sample. Within those guesses, there typically exists at least 1 
        face but can be more. By default, it should be 43,008.
    
    Be default, this class initializes the following attributes which 
    can be changed after initialization of the class (but, typically, 
    should not be changed):

    Attributes:
        nms_threshold (float): The threshold, based on which 
            multiple bounding box or landmark predictions for the same 
            face are merged into one. Defaults to 0.4.
        variance (list[int]): The variance of the bounding boxes 
            used to undo the encoding of coordinates of raw  bounding 
            box and landmark predictions.
    """
    #: WEIGHTS_FILENAME (str): The constant specifying the name of 
    #: ``.pth`` file from which the weights for this model should be 
    #: loaded. Defaults to "retinaface_detector.pth".
    WEIGHTS_FILENAME = "retinaface_detector.pth"

    def __init__(self, strategy: str = "all", vis: float = 0.6):
        """Initializes RetinaFace model.            
        
        This method initializes ResNet-50 backbone and further 
        layers required for face detection and bbox/landm predictions.

        Args:
            strategy: The strategy used to retrieve the landmarks when
                :meth:`predict` is called. The available options are:

                    * "all" - landmarks for all faces per single image
                      (single batch entry) will be considered.
                    * "best" - landmarks for a single face with the
                      highest confidence score per image will be 
                      considered.
                    * "largest" - landmarks for a single largest face
                      per image will be considered.

                The most efficient option is 'best' and the least
                efficient is "largest". Defaults to "all".
            vis: The visual threshold, i.e., minimum confidence score,
                for a face to be considered an actual face. Lower
                scores will allow the detection of more faces per image
                but can result in non-actual faces, e.g., random
                surfaces somewhat representing faces. Higher scores will 
                prevent detecting faulty faces but may result in only a
                few faces detected, whereas there can be more, e.g., 
                higher will prevent the detection of blurry faces. 
                Defaults to 0.6.
        """
        super().__init__()

        # Initialize attributes
        self.strategy = strategy
        self.vis_threshold = vis
        self.nms_threshold = 0.4
        self.variance = [0.1, 0.2]

        # Set up backbone and config
        backbone = models.resnet50()
        in_channels, out_channels = 256, 256
        in_channels_list = [in_channels * x for x in [2, 4, 8]]
        return_layers = {'layer2': 1, 'layer3': 2, 'layer4': 3}

        # Construct the backbone by retrieving intermediate layers
        self.body = _utils.IntermediateLayerGetter(backbone, return_layers)

        # Construct sub-layers to extract features for heads
        self.fpn = FPN(in_channels_list, out_channels)
        self.ssh1 = SSH(out_channels, out_channels)
        self.ssh2 = SSH(out_channels, out_channels)
        self.ssh3 = SSH(out_channels, out_channels)

        # Construct 3 heads - score, bboxes & landms
        self.ClassHead = Head.make(2, out_channels)
        self.BboxHead = Head.make(4, out_channels)
        self.LandmarkHead = Head.make(10, out_channels)

    def forward(
        self,
        x: torch.Tensor,
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """Performs forward pass.

        Takes an input batch and performs inference based on the modules 
        it has. Returns an unfiltered tuple of scores, bounding boxes 
        and landmarks for all the possible detected faces. The 
        predictions are encoded to comfortably compute the loss during 
        training and thus should be decoded to coordinates.

        Args:
            x: The input tensor of shape (N, 3, H, W).

        Returns:
            A tuple of torch tensors where the first element is
            confidence scores for each prediction of shape
            (N, out_dim, 2) with values between 0 and 1 representing
            probabilities, the second element is bounding boxes of shape 
            (N, out_dim, 4) with unbounded values and the last element 
            is landmarks of shape (N, ``out_dim``, 10) with unbounded
            values.
        """
        # Extract FPN + SSH features
        fpn = self.fpn(self.body(x))
        fts = [self.ssh1(fpn[0]), self.ssh2(fpn[1]), self.ssh3(fpn[2])]

        # Create head list and use each to process feature list
        hs = [self.ClassHead, self.BboxHead, self.LandmarkHead]
        pred = [torch.cat([h[i](f) for i, f in enumerate(fts)], 1) for h in hs]
        
        return F.softmax(pred[0], dim=-1), pred[1], pred[2]
    
    def decode_bboxes(
        self,
        loc: torch.Tensor,
        priors: torch.Tensor,
    ) -> torch.Tensor:
        """Decodes bounding boxes from predictions.

        Takes the predicted bounding boxes (locations) and undoes the 
        encoding for offset regression used at training time.

        Args:
            loc: Bounding box (location) predictions for loc layers of
                shape (N, out_dim, 4). 
            priors: Prior boxes in center-offset form of shape
                (out_dim, 4).

        Returns:
            A tensor of shape (N, out_dim, 4) representing decoded
            bounding box predictions where the last dim can be
            interpreted as x1, y1, x2, y2 coordinates - the start and
            the end corners defining the face box.
        """
        # Concatenate priors
        boxes = torch.cat((
            priors[:, :2] + loc[..., :2] * self.variance[0] * priors[:, 2:],
            priors[:, 2:] * torch.exp(loc[..., 2:] * self.variance[1])
        ), 2)
        
        # Adjust values for proper xy coords
        boxes[..., :2] -= boxes[..., 2:] / 2
        boxes[..., 2:] += boxes[..., :2]

        return boxes

    def decode_landms(
        self,
        pre: torch.Tensor,
        priors: torch.Tensor,
    ) -> torch.Tensor:
        """Decodes landmarks from predictions.

        Takes the predicted landmarks (pre) and undoes the encoding for
        offset regression used at training time.

        Args:
            pre: Landmark predictions for loc layers of shape
                (N, out_dim, 10).
            priors: Prior boxes in center-offset form of shape
                (out_dim, 4).

        Returns:
            A tensor of shape (N, out_dim, 10) representing decoded
            landmark predictions where the last dim can be
            interpreted as x1, y1, ..., x10, y10 coordinates - one for 
            each of the 5 landmarks.
        """
        # Concatenate priors
        var = self.variance
        landms = torch.cat((
            priors[..., :2] + pre[..., :2] * var[0] * priors[..., 2:],
            priors[..., :2] + pre[..., 2:4] * var[0] * priors[..., 2:],
            priors[..., :2] + pre[..., 4:6] * var[0] * priors[..., 2:],
            priors[..., :2] + pre[..., 6:8] * var[0] * priors[..., 2:],
            priors[..., :2] + pre[..., 8:10] * var[0] * priors[..., 2:],
        ), dim=2)

        return landms
    
    def filter_preds(
        self,
        scores: torch.Tensor,
        bboxes: torch.Tensor,
        landms: torch.Tensor,
    ) -> tuple[torch.Tensor, torch.Tensor, list[int]]:
        """Filters predictions for identified faces for each sample.
        
        This method works as follows:

            1. First, it filters out bad predictions based on
               ``self.vis_threshold``.
            2. Then it gathers all the remaining predictions across the
               batch dimension, i.e., the batch dimension becomes not
               the number of samples but the number of filtered out
               predictions.
            3. It loops for each set of filtered predictions per sample
               sorting each set of confidence scores from best to worst.
            4. For each set of confidence scores, it identifies distinct 
               faces and keeps the record of which indices to keep. At 
               this stage it uses ``self.nms_threshold`` to remove the 
               duplicate face predictions.
            5. Finally, it applies the kept indices for each person
               (each face) to select corresponding bounding boxes and
               landmarks.

        Args:
            scores: The confidence score predictions of shape
                (N, out_dim).
            bboxes: The bounding boxes for each face of shape
                (N, out_dim, 4) where the last 4 numbers correspond to
                start and end coordinates - x1, y1, x2, y2.
            landms: The landmarks for each face of shape
                (N, out_dim, num_landmarks * 2) where the last dim 
                corresponds to landmark coordinates x1, y1, ... . By
                default, num_landmarks is 5.

        Returns:
            A tuple where the first element is a torch tensor of shape
            (``num_faces``, 4), the second element is a torch tensor of
            shape (``num_faces``, ``num_landmarks`` * 2) and the third 
            element is a list of length ``num_faces``. First and second 
            elements correspond to bounding boxes and landmarks for each 
            face across all samples and the third element provides an 
            index for each bounding box/set of landmarks that identifies
            which sample that box/set (or that face) is extracted from
            (because each sample can have multiple faces).
        """
        # Init variables, identify masks to filter best faces
        cumsum, people_indices, sample_indices = 0, [], []
        masks = scores > self.vis_threshold

        # Flatten across batch filtered predictions, compute face areas
        scores, bboxes, landms = scores[masks], bboxes[masks], landms[masks]
        areas = (bboxes[:, 2]-bboxes[:, 0]+1) * (bboxes[:, 3]-bboxes[:, 1]+1)

        for i, num_valid in enumerate(masks.sum(dim=1)):
            # Extract all face preds for a single sample
            start, end, keep = cumsum, cumsum+num_valid, []
            bbox, area = bboxes[start:end], areas[start:end]
            scores_sorted = scores[start:end].argsort(descending=True)

            while scores_sorted.numel() > 0:
                # Append best face's index to keep
                keep.append(j := scores_sorted[0])
                
                # Find coordinates that at least bound the current face
                xy1 = torch.maximum(bbox[j, :2], bbox[scores_sorted[1:], :2])
                xy2 = torch.minimum(bbox[j, 2:], bbox[scores_sorted[1:], 2:])

                # Compute width and height for the current minimal face
                w = torch.maximum(torch.tensor(0.0), xy2[:, 0] - xy1[:, 0] + 1)
                h = torch.maximum(torch.tensor(0.0), xy2[:, 1] - xy1[:, 1] + 1)

                # Compute nms for identifying areas for the current face
                ovr = (a := w * h) / (area[j] + area[scores_sorted[1:]] - a)
                
                # Filter out current face, keep next best scores
                inds = torch.where(ovr <= self.nms_threshold)[0]
                scores_sorted = scores_sorted[inds + 1]
            
            # Update people and sample indices, increment cumsum
            people_indices.extend([cumsum + k for k in keep])
            sample_indices.extend([i] * len(keep))
            cumsum += num_valid
        
        # Select the final landms and bboxes
        bboxes = bboxes[people_indices, :]
        landms = landms[people_indices, :]
        
        return landms, bboxes, sample_indices
    
    def take_by_strategy(
        self,
        landms: torch.Tensor,
        bboxes: torch.Tensor,
        idx: list[int],
    ) -> tuple[torch.Tensor, list[int]]:
        """Filters landmarks according to strategy.

        This method takes a batch of landmarks and bounding boxes (one
        for each face) filters only specific landmarks by a specific
        strategy. Here are the following cases of strategy:

            * "all" - effectively, nothing is done and simply the
              already passed values `landms` and `idx` are returned 
              without any changes.
            * "best" - the very first set of landmarks for each image 
              image is returned (the first set is the best set because
              the landmarks were sorted when duplicates were filtered
              out in :meth:`filter_preds`). This means
              the returned indices list is unique, e.g., goes from 
              ``[0, 0, 0, 1, 1, 2, 3, 3]`` to ``[0, 1, 2, 3]``.
            * "largest" - similar to 'best', except that this strategy
              requires performing additional computation to find out the 
              largest face based on the area of bounding boxes. Thus the 
              length of the `idx` list (which is equal to the number of 
              sets of landmarks) is the same as for 'best' strategy,
              except not the first (best) faces (actually, their
              landmarks) for each image but selected faces are returned.

        Note:
            Strategy "best" is most memory efficient, strategy "largest" 
            is least time efficient. Strategy "all" is as fast as "best" 
            but takes up more space.

        Args:
            landms: Landmarks batch of shape 
                (``num_faces``, ``num_landm`` * 2).
            bboxes: Bounding boxes batch of shape (``num_faces``, 4).
            idx: Indices where each index maps to an image from
                which some face prediction (landmarks and bounding box) 
                was retrieved. For instance if the 2nd element of idx is 
                1, that means that the 2nd element of ``landms`` and the
                2nd element of ``bboxes`` correspond to the 1st image. 
                This list is ascending, meaning the elements are
                grouped and increase, for example, the list may look
                like this: ``[0, 0, 1, 2, 3, 3, 3, 3, 4, 4, 5, 6, 6]``.

        Raises:
            ValueError: If the strategy is not supported.

        Returns:
            A tuple where the first element is torch tensor of shape 
            (``num_faces``, ``num_landm`` * 2) representing the selected 
            sets of landmarks and the second element is a list of 
            indices where each index maps a corresponding set of 
            landmarks (face) to an image identified by that index.
        """
        if len(idx) == 0:
            # If no predicted landmarks, return empty lists
            return torch.tensor([], device=landms.device), []
        
        # Init helper variables
        landmarks, indices = [], []
        cache = {"idx": [], "bboxes": [], "landms": []}

        for i in range(len(idx)):
            # Apend everything to cache
            cache["idx"].append(idx[i])
            cache["bboxes"].append(bboxes[i])
            cache["landms"].append(landms[i])

            if i != len(idx) - 1 and cache["idx"][-1] == idx[i + 1]:
                # No operations until cache for current idx is full
                continue

            match self.strategy:
                case "all":
                    # Append all landmarks and indices
                    landmarks.extend(cache["landms"])
                    indices.extend(cache["idx"])
                case "best":
                    # Append the first set of landmarks
                    landmarks.append(cache["landms"][0])
                    indices.append(cache["idx"][0])
                case "largest":
                    # Compute bounding box areas
                    bbs = torch.stack(cache["bboxes"])
                    areas = (bbs[:, 2] - bbs[:, 0] + 1) *\
                            (bbs[:, 3] - bbs[:, 1] + 1)

                    # Append only the largest face landmarks and its idx
                    landmarks.append(cache["landms"][areas.argmax()])
                    indices.append(cache["idx"][0])
                case _:
                    raise ValueError(f"Unsupported startegy: {self.strategy}")
            
            # Clear cache (reinitialize empty lists)
            cache = {k: [] for k in cache.keys()}

        # Stack landmarks across batch dim
        landmarks = torch.stack(landmarks)
    
        return landmarks, indices
    
    @torch.no_grad()
    def predict(self, images: torch.Tensor) -> tuple[np.ndarray, list[int]]:
        """Predict the sets of landmarks from the image batch.

        This method takes a batch of images, detect all visible faces, 
        predicts bounding boxes and landmarks for each face and then 
        filters those faces according to a specific strategy - see
        :meth:`take_by_strategy` for more info. Finally, it returns 
        those selected sets of landmarks and corresponding indices that 
        map each set to a specific image where the face was originally 
        detected.

        The predicted sets of landmarks are 5-point coordinates (they  
        are specified from an observer's viewpoint, meaning that, for 
        instance, left eye is the eye on the left hand-side of the image 
        rather than the left eye from the person's to whom the eye 
        belongs perspective):

            1. **(x1, y1)** - coordinate of the left eye
            2. **(x2, y2)** - coordinate of the right eye
            3. **(x3, y3)** - coordinate of the nose tip
            4. **(x4, y4)** - coordinate of the left mouth corner
            5. **(x5, y5)** - coordinate of the right mouth corner

        The coordinates are with respect to the sizes of the images 
        (typically padded) provided as an input to this method.

        Args:
            images: Image batch of shape (N, 3, H, W) in RGB form with 
                float values from 0.0 to 255.0. It must be on the same 
                device as this model.

        Returns:
            A tuple where the first element is a numpy array of shape 
            (``num_faces``, 5, 2) representing the selected sets of 
            landmark coordinates and the second element is a list of
            corresponding indices mapping each face to an image it comes
            from.
        """
        # Convert images to appropriate input and perform inference
        x, offset = images[:, [2, 1, 0]], torch.tensor([104, 117, 123])
        scores, bboxes, landms = self(x - offset.view(3, 1, 1).to(x.device))

        # Create prior boxes and scale factors to decode bboxes & landms
        priors = PriorBox((x.size(2), x.size(3))).forward().to(x.device)
        scale_b = torch.tensor([x.size(3), x.size(2)] * 2, device=x.device)
        scale_l = torch.tensor([x.size(3), x.size(2)] * 5, device=x.device)

        # Decode the predictions
        scores = scores[..., 1]
        bboxes = self.decode_bboxes(bboxes, priors) * scale_b
        landms = self.decode_landms(landms, priors) * scale_l

        # Filter out bad predictions, then filter by strategy
        filtered = self.filter_preds(scores, bboxes, landms)
        landmarks, indices = self.take_by_strategy(*filtered)

        # Stack landmarks across batch dim and reshape as coords
        landmarks = landmarks.view(-1, 5, 2).cpu().numpy()

        return landmarks, indices



import torch
import torch.nn as nn
import torch.nn.functional as F

from math import ceil
from itertools import product

########################################################################
############################### GENERAL ################################
########################################################################

class LoadMixin():
    URL_ROOT = "https://github.com/mantasu/face-crop-plus/releases/download/v1.0.0/"
    WEIGHTS_FILENAME = None

    def load(self, device: str | torch.device = "cpu"):
        weights = self.get_weights(device)
        self.load_state_dict(weights)
        self.to(device)
        self.eval()
        
        for param in self.parameters():
            param.requires_grad = False
        
        return self
    
    def get_weights(self, device):
        if self.WEIGHTS_FILENAME is None:
            raise ValueError(f"Please ensure 'WEIGHTS_FILENAME' is specified "
                              "for the class that inherits this mixin.")
        
        url = self.URL_ROOT + self.WEIGHTS_FILENAME
        weights = torch.hub.load_state_dict_from_url(url, map_location=device)

        return weights

########################################################################
######################### RETINA FACE MODULES ##########################
########################################################################

class PriorBox():
    def __init__(self, size):
        super().__init__()
        self.size = size
        self.steps = [8, 16, 32]
        self.min_sizes = [[16, 32], [64, 128], [256, 512]]
        self.features = [tuple(ceil(x / n) for x in size) for n in self.steps]

    def forward(self):
        anchors = []
        
        for k, f in enumerate(self.features):
            min_sizes = self.min_sizes[k]

            for i, j in product(range(f[0]), range(f[1])):
                for min_size in min_sizes:
                    a, b = min_size / self.size[1], min_size / self.size[0]
                    dx = [x * self.steps[k] / self.size[1] for x in [j + 0.5]]
                    dy = [y * self.steps[k] / self.size[0] for y in [i + 0.5]]
                    anchors.extend([(x, y, a, b) for y, x in product(dy, dx)])
        
        return torch.tensor(anchors).view(-1, 4)

class SSH(nn.Module):
    def __init__(self, in_channel, out_channel):
        super().__init__()

        leaky = 0.1 if out_channel <= 64 else 0
        out_dim = out_channel // 4

        self.conv3X3 = self.conv_bn_no_relu(in_channel, out_channel // 2)
        self.conv5X5_1 = self.conv_bn(in_channel, out_dim, leaky = leaky)
        self.conv5X5_2 = self.conv_bn_no_relu(out_dim, out_dim)
        self.conv7X7_2 = self.conv_bn(out_dim, out_dim, leaky = leaky)
        self.conv7x7_3 = self.conv_bn_no_relu(out_dim, out_dim)
    
    def conv_bn(self, inp, oup, stride=1, leaky=0):
        return nn.Sequential(
            nn.Conv2d(inp, oup, 3, stride, 1, bias=False),
            nn.BatchNorm2d(oup),
            nn.LeakyReLU(negative_slope=leaky, inplace=True)
        )

    def conv_bn_no_relu(self, inp, oup, stride=1):
        return nn.Sequential(
            nn.Conv2d(inp, oup, 3, stride, 1, bias=False),
            nn.BatchNorm2d(oup),
        )

    def forward(self, input):
        conv3X3 = self.conv3X3(input)
        conv5X5_1 = self.conv5X5_1(input)
        conv5X5 = self.conv5X5_2(conv5X5_1)
        conv7X7_2 = self.conv7X7_2(conv5X5_1)
        conv7X7 = self.conv7x7_3(conv7X7_2)

        return F.relu(torch.cat([conv3X3, conv5X5, conv7X7], dim=1))

class FPN(nn.Module):
    def __init__(self, in_channels_list, out_channels):
        super().__init__()

        leaky = 0.1 if out_channels <= 64 else 0
        out = out_channels

        self.output1 = self.conv_bn1X1(in_channels_list[0], out, leaky = leaky)
        self.output2 = self.conv_bn1X1(in_channels_list[1], out, leaky = leaky)
        self.output3 = self.conv_bn1X1(in_channels_list[2], out, leaky = leaky)

        self.merge1 = self.conv_bn(out, out, leaky = leaky)
        self.merge2 = self.conv_bn(out, out, leaky = leaky)
    
    def conv_bn(self, inp, oup, stride = 1, leaky = 0):
        return nn.Sequential(
            nn.Conv2d(inp, oup, 3, stride, 1, bias=False),
            nn.BatchNorm2d(oup),
            nn.LeakyReLU(negative_slope=leaky, inplace=True)
        )
    
    def conv_bn1X1(self, inp, oup, stride=1, leaky=0):
        return nn.Sequential(
            nn.Conv2d(inp, oup, 1, stride, padding=0, bias=False),
            nn.BatchNorm2d(oup),
            nn.LeakyReLU(negative_slope=leaky, inplace=True)
        )

    def forward(self, input):
        input = list(input.values())

        output1 = self.output1(input[0])
        output2 = self.output2(input[1])
        output3 = self.output3(input[2])

        size_up3 = [output2.size(2), output2.size(3)]
        size_up2 = [output1.size(2), output1.size(3)]

        up3 = F.interpolate(output3, size=size_up3, mode="nearest")
        output2 = output2 + up3
        output2 = self.merge2(output2)

        up2 = F.interpolate(output2, size=size_up2, mode="nearest")
        output1 = output1 + up2
        output1 = self.merge1(output1)

        return [output1, output2, output3]

class Head(nn.Module):
    def __init__(self, num_out, in_channels=512):
        super().__init__()
        self.num_out = num_out
        self.conv1x1 = nn.Conv2d(in_channels, 2 * num_out, 1)
    
    def forward(self, x):
        x = self.conv1x1(x)
        out = x.permute(0, 2, 3, 1).contiguous()
        
        return out.view(out.size(0), -1, self.num_out)
    
    @classmethod
    def make(cls, num_out, in_channels=512):
        head = nn.ModuleList([cls(num_out, in_channels) for _ in range(3)])
        return head

########################################################################
############################# RRDB MODULES #############################
########################################################################

class ResidualDenseBlock_5C(nn.Module):
    def __init__(self, nf=64, gc=32, bias=True):
        super().__init__()

        self.conv1 = nn.Conv2d(nf, gc, 3, 1, 1, bias=bias)
        self.conv2 = nn.Conv2d(nf + gc, gc, 3, 1, 1, bias=bias)
        self.conv3 = nn.Conv2d(nf + 2 * gc, gc, 3, 1, 1, bias=bias)
        self.conv4 = nn.Conv2d(nf + 3 * gc, gc, 3, 1, 1, bias=bias)
        self.conv5 = nn.Conv2d(nf + 4 * gc, nf, 3, 1, 1, bias=bias)
        self.lrelu = nn.LeakyReLU(negative_slope=0.2, inplace=True)

    def forward(self, x):
        x1 = self.lrelu(self.conv1(x))
        x2 = self.lrelu(self.conv2(torch.cat((x, x1), 1)))
        x3 = self.lrelu(self.conv3(torch.cat((x, x1, x2), 1)))
        x4 = self.lrelu(self.conv4(torch.cat((x, x1, x2, x3), 1)))
        x5 = self.conv5(torch.cat((x, x1, x2, x3, x4), 1))

        return x5 * 0.2 + x

class RRDB(nn.Module):
    def __init__(self, nf, gc=32):
        super().__init__()
        self.RDB1 = ResidualDenseBlock_5C(nf, gc)
        self.RDB2 = ResidualDenseBlock_5C(nf, gc)
        self.RDB3 = ResidualDenseBlock_5C(nf, gc)

    def forward(self, x):
        out = self.RDB1(x)
        out = self.RDB2(out)
        out = self.RDB3(out)

        return out * 0.2 + x

########################################################################
############################# BISE MODULES #############################
########################################################################

class BasicBlock(nn.Module):
    def __init__(self, in_chan, out_chan, stride=1):
        super().__init__()

        self.conv1 = self.conv3x3(in_chan, out_chan, stride)
        self.bn1 = nn.BatchNorm2d(out_chan)
        self.conv2 = self.conv3x3(out_chan, out_chan)
        self.bn2 = nn.BatchNorm2d(out_chan)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = None

        if in_chan != out_chan or stride != 1:
            self.downsample = nn.Sequential(
                nn.Conv2d(in_chan, out_chan, 1, stride=stride, bias=False),
                nn.BatchNorm2d(out_chan),
            )
    
    def conv3x3(self, in_planes, out_planes, stride=1):
        return nn.Conv2d(in_planes, out_planes, 3, stride, 1, bias=False)

    def forward(self, x):
        residual = self.conv1(x)
        residual = F.relu(self.bn1(residual))
        residual = self.conv2(residual)
        residual = self.bn2(residual)
        shortcut = x

        if self.downsample is not None:
            shortcut = self.downsample(x)

        out = shortcut + residual
        out = self.relu(out)

        return out

class Resnet18(nn.Module):
    def __init__(self):
        super().__init__()

        self.conv1 = nn.Conv2d(3, 64, 7, stride=2, padding=3, bias=False)
        self.bn1 = nn.BatchNorm2d(64)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        self.layer1 = self.create_layer_basic(64, 64, bnum=2, stride=1)
        self.layer2 = self.create_layer_basic(64, 128, bnum=2, stride=2)
        self.layer3 = self.create_layer_basic(128, 256, bnum=2, stride=2)
        self.layer4 = self.create_layer_basic(256, 512, bnum=2, stride=2)
    
    def create_layer_basic(self, in_chan, out_chan, bnum, stride=1):
        layers = [BasicBlock(in_chan, out_chan, stride=stride)]

        for _ in range(bnum-1):
            layers.append(BasicBlock(out_chan, out_chan, stride=1))

        return nn.Sequential(*layers)

    def forward(self, x):
        x = self.conv1(x)
        x = F.relu(self.bn1(x))
        x = self.maxpool(x)

        x = self.layer1(x)
        feat8 = self.layer2(x)
        feat16 = self.layer3(feat8)
        feat32 = self.layer4(feat16)

        return feat8, feat16, feat32

class ConvBNReLU(nn.Module):
    def __init__(self, in_ch, out_ch, ks=3, stride=1, padding=1):
        super().__init__()
        self.conv = nn.Conv2d(in_ch, out_ch, ks, stride, padding, bias=False)
        self.bn = nn.BatchNorm2d(out_ch)

    def forward(self, x):
        x = self.conv(x)
        x = F.relu(self.bn(x))

        return x

class BiSeNetOutput(nn.Module):
    def __init__(self, in_chan, mid_chan, n_classes):
        super().__init__()
        self.conv = ConvBNReLU(in_chan, mid_chan, ks=3, stride=1, padding=1)
        self.conv_out = nn.Conv2d(mid_chan, n_classes, 1, bias=False)

    def forward(self, x):
        x = self.conv(x)
        x = self.conv_out(x)

        return x

class AttentionRefinementModule(nn.Module):
    def __init__(self, in_chan, out_chan):
        super().__init__()
        self.conv = ConvBNReLU(in_chan, out_chan, ks=3, stride=1, padding=1)
        self.conv_atten = nn.Conv2d(out_chan, out_chan, 1, bias=False)
        self.bn_atten = nn.BatchNorm2d(out_chan)
        self.sigmoid_atten = nn.Sigmoid()

    def forward(self, x):
        feat = self.conv(x)
        atten = F.avg_pool2d(feat, feat.size()[2:])
        atten = self.conv_atten(atten)
        atten = self.bn_atten(atten)
        atten = self.sigmoid_atten(atten)
        out = torch.mul(feat, atten)

        return out

class ContextPath(nn.Module):
    def __init__(self):
        super().__init__()

        self.resnet = Resnet18()
        self.arm16 = AttentionRefinementModule(256, 128)
        self.arm32 = AttentionRefinementModule(512, 128)
        self.conv_head32 = ConvBNReLU(128, 128, ks=3, stride=1, padding=1)
        self.conv_head16 = ConvBNReLU(128, 128, ks=3, stride=1, padding=1)
        self.conv_avg = ConvBNReLU(512, 128, ks=1, stride=1, padding=0)

    def forward(self, x):
        feat8, feat16, feat32 = self.resnet(x)
        H8, W8 = feat8.size()[2:]
        H16, W16 = feat16.size()[2:]
        H32, W32 = feat32.size()[2:]

        avg = F.avg_pool2d(feat32, feat32.size()[2:])
        avg = self.conv_avg(avg)
        avg_up = F.interpolate(avg, (H32, W32))

        feat32_arm = self.arm32(feat32)
        feat32_sum = feat32_arm + avg_up
        feat32_up = F.interpolate(feat32_sum, (H16, W16))
        feat32_up = self.conv_head32(feat32_up)

        feat16_arm = self.arm16(feat16)
        feat16_sum = feat16_arm + feat32_up
        feat16_up = F.interpolate(feat16_sum, (H8, W8))
        feat16_up = self.conv_head16(feat16_up)

        return feat8, feat16_up

class FeatureFusionModule(nn.Module):
    def __init__(self, in_chan, out_chan):
        super().__init__()
        self.convblk = ConvBNReLU(in_chan, out_chan, ks=1, stride=1, padding=0)
        self.conv1 = nn.Conv2d(out_chan, out_chan // 4, 1, bias=False)
        self.conv2 = nn.Conv2d(out_chan // 4, out_chan, 1, bias=False)
        self.relu = nn.ReLU(inplace=True)
        self.sigmoid = nn.Sigmoid()

    def forward(self, fsp, fcp):
        fcat = torch.cat([fsp, fcp], dim=1)
        feat = self.convblk(fcat)
        atten = F.avg_pool2d(feat, feat.size()[2:])
        atten = self.conv1(atten)
        atten = self.relu(atten)
        atten = self.conv2(atten)
        atten = self.sigmoid(atten)
        feat_atten = torch.mul(feat, atten)
        feat_out = feat_atten + feat

        return feat_out



import setuptools

with open("README.md", "r", encoding = "utf-8") as f:
    long_description = f.read()

setuptools.setup(
    name = "face-crop-plus",
    version = "1.0.3",
    author = "Mantas Birškus",
    author_email = "mantix7@gmail.com",
    license = "MIT",
    description = f"Automatic face aligner and cropper with quality "
                  f"enhancement and attribute parsing.",
    long_description = long_description,
    long_description_content_type = "text/markdown",
    url = "https://github.com/mantasu/face-crop-plus",
    project_urls = {
        "Documentation": "https://mantasu.github.io/face-crop-plus",
        "Bug Tracker": "https://github.com/mantasu/face-crop-plus/issues",
    },
    keywords = [
        "face",
        "python",
        "pytorch",
        "alignment",
        "cropping",
        "super resolution",
        "quality enhancement",
        "parsing",
        "grouping",
        "attributes",
        "mask",
        "segmentation",
        "celeba",
    ],
    install_requires = [
        "tqdm",
        "opencv-python",
        "torch",
        "torchvision",
    ],
    classifiers = [
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    package_dir = {"": "src"},
    packages = setuptools.find_packages(where="src"),
    entry_points={
        "console_scripts": [
            "face-crop-plus=face_crop_plus.__main__:main",
            "face_crop_plus=face_crop_plus.__main__:main",
            "fcp=face_crop_plus.__main__:main",
        ]
    },
    python_requires = ">=3.10"
)